/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

const nsISupports            = Components.interfaces.nsISupports;

const nsICategoryManager     = Components.interfaces.nsICategoryManager;
const nsIComponentRegistrar  = Components.interfaces.nsIComponentRegistrar;
const nsIContentHandler      = Components.interfaces.nsIContentHandler;
const nsIFactory             = Components.interfaces.nsIFactory;
const nsIModule              = Components.interfaces.nsIModule;

const nsIChannel             = Components.interfaces.nsIChannel;
const nsIDOMWindow           = Components.interfaces.nsIDOMWindow;
const nsIFile                = Components.interfaces.nsIFile;
const nsIFileProtocolHandler = Components.interfaces.nsIFileProtocolHandler;
const nsIIOService           = Components.interfaces.nsIIOService;
const nsIProperties          = Components.interfaces.nsIProperties;
const nsIRDFContainer        = Components.interfaces.nsIRDFContainer;
const nsIRDFLiteral          = Components.interfaces.nsIRDFLiteral;
const nsIRDFResource         = Components.interfaces.nsIRDFResource;
const nsIRDFService          = Components.interfaces.nsIRDFService;
const nsIWindowMediator      = Components.interfaces.nsIWindowMediator;
const nsIBrowserHistory      = Components.interfaces.nsIBrowserHistory;
const nsISHistoryInternal    = Components.interfaces.nsISHistoryInternal;
const nsIXULAppInfo          = Components.interfaces.nsIXULAppInfo;

const nsIMIMEService         = Components.interfaces.nsIMIMEService;

/**
 * 本体
 * message/rfc822 等を開く
 *   Inherits From: nsIContentHandler
 */
var arUnMHTContentHandler = {
    /**
     * インターフェースの要求
     *   nsISupports.QueryInterface
     *
     * @param  nsIIDRef iid
     *         インターフェース ID
     * @throws Components.results.NS_NOINTERFACE
     * @return nsIContentHandler
     *         this
     */
    QueryInterface : function (iid) {
        if (iid.equals (nsISupports)
            || iid.equals (nsIContentHandler)) {
            return this;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
    
    /**
     * mht ファイルを処理する
     *   nsIContentHandler.handleContent
     *
     * @param  String contentType
     *         MIME-Type
     * @param  Window context
     *         参照元
     * @param  nsIRequest request
     *         リクエスト
     *
     * 旧バージョンでは
     * @param  String contentType
     *         MIME-Type
     * @param  String context
     *         コマンド (command/不明)
     * @param  Window request
     *         参照元 (context)
     * @param  nsIRequest dummy
     *         リクエスト (request)
     */
    handleContent : function (contentType, context, request, dummy) {
        var targetWindow = null;
        try {
            targetWindow = context.getInterface (nsIDOMWindow);
        }
        catch (e) {
            /* 旧バージョンの場合 */
            
            /* context 以降の引数を1つずらす */
            context = request;
            request = dummy;
            
            targetWindow = context;
        }
        var channel = null;
        var uri = null;
        try {
            channel = request.QueryInterface (nsIChannel);
            uri = channel.URI;
        }
        catch (e) {
        }
        if (uri) {
            if (uri.scheme == "imap"
                || uri.scheme == "mailbox"
                || uri.scheme == "news"
                || uri.scheme == "nntp"
                || uri.scheme == "pop"
                || uri.scheme == "nsews") {
                var NS_ERROR_WONT_HANDLE_CONTENT = 0x805d0001;
                throw NS_ERROR_WONT_HANDLE_CONTENT;
            }
            if (contentType == "application/octet-stream"
                || contentType == "application/x-maf") {
                if (!uri.spec.match (/[^\/]\.mht$/)) {
                    var NS_ERROR_WONT_HANDLE_CONTENT = 0x805d0001;
                    throw NS_ERROR_WONT_HANDLE_CONTENT;
                }
            }
            var mediator
            = Components.classes ["@mozilla.org/appshell/window-mediator;1"]
            .getService (nsIWindowMediator);
            var chromeWindow
            = mediator.getMostRecentWindow ("navigator:browser");
            
            try {
                var browserHistory
                    = Components.classes
                    ["@mozilla.org/browser/global-history;2"]
                    .getService (nsIBrowserHistory);
                browserHistory.addURI (uri, false, true, null);
            }
            catch (e) {
                try {
                    var browserHistory
                    = Components.classes
                    ["@mozilla.org/browser/global-history;1"]
                    .getService (nsIBrowserHistory);
                    browserHistory.addPage (uri.spec);
                }
                catch (e) {
                }
            }
            
            chromeWindow.UnMHT.loadMHT (uri, targetWindow,
                                        false,  /* ドキュメント無し */
                                        false,  /* 履歴はそのまま */
                                        0,      /* 展開は設定に任せる */
                                        "",     /* ルートドキュメントを開く */
                                        false); /* キャッシュは使用しない */
        }
    }
};

/**
 * 本体のファクトリー
 *   Inherits From: nsIFactory
 */
var arUnMHTContentHandlerFactory = {
    /**
     * 本体を生成する
     *   nsIFactory.createInstance
     *
     * @param  nsISupport outer
     *          統合する対象
     * @param  nsIIDRef iid
     *         生成する対象のインターフェース ID
     * @return arUnMHTContentHandler
     *         本体
     */
    createInstance : function (outer, iid) {
        if (outer != null) {
            /* 統合する対象がある場合はエラー */
            throw Components.results.NS_ERROR_NO_AGGREGATION;
        }
        
        return arUnMHTContentHandler.QueryInterface (iid);
    }
};

/**
 * XPCOM のモジュール
 *   Inherits From: nsIModule
 */
var arUnMHTContentHandlerModule = {
    /* 本体に関する情報 */
    CONTRACTID: "@unmht.org/unmht-content-handler;1",
    CID: Components.ID ("{4d55009b-81cf-4169-af2d-c7c3ed13af4f}"),
    CNAME: "UnMHT Content Handler JS Component",
    
    /**
     * インターフェースの要求
     *   nsISupports.QueryInterface
     *
     * @param  nsIIDRef iid
     *         インターフェースID
     * @throws Components.results.NS_NOINTERFACE
     * @return nsIModule
     *         this
     */
    QueryInterface : function (iid) {
        if (iid.equals (nsISupports)
            || iid.equals (nsIModule)) {
            return this;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
    
    /**
     * 登録処理
     *   nsIModule.registerSelf
     *
     * @param  nsIComponentManager compMgr
     * @param  nsIFile fileSpec
     *         モジュールのファイル
     *           reference と他のソースコード中で並びが違う
     * @param  String location
     *         不明
     *           reference と他のソースコード中で並びが違う
     * @param  String type
     *         ローダの種類
     */
    registerSelf : function (compMgr, fileSpec, location, type) {
        try {
            var id
            = Components.classes ["@mozilla.org/xre/app-info;1"]
            .getService (nsIXULAppInfo).ID;
            if (id == "{3550f703-e582-4d05-9a08-453d09bdfdc6}") {
                /* Thunderbird */
                return;
            }
        } catch (e) {}
        
        compMgr = compMgr.QueryInterface (nsIComponentRegistrar);
        compMgr.registerFactoryLocation (this.CID,
                                         this.CNAME,
                                         this.CONTRACTID,
                                         fileSpec, location, type);
        
        var prefix = "@mozilla.org/uriloader/content-handler;1?type=";
        
        var types = [
            "application/octet-stream",
            "message/rfc822",
            "application/mime",
            "application/x-mht",
            "text/x-mht"
            ];
        
        try {
            var mime = Components.classes ["@mozilla.org/mime;1"]
                .getService (nsIMIMEService);
            var systemDefined = mime.getTypeFromExtension ("mht");
            var exists = false;
        
            for (var i = 0; i < types.length; i ++) {
                if (systemDefined == types [i]) {
                    exists = true;
                    break;
                }
            }
        
            if (!exists) {
                types.push (systemDefined);
            }
        }
        catch (e) {
        }
        
        /* 各ユーザ毎の mimeTypes.rdf をチェックする */
        try {
            var profdir
                = Components
                .classes ["@mozilla.org/file/directory_service;1"]
                .getService (nsIProperties)
                .get ("DefProfRt", nsIFile);
            var entries = profdir.directoryEntries;
            while (entries.hasMoreElements ()) {
                var file = entries.getNext ();
                try {
                    file = file.QueryInterface (nsIFile);
                    if (file.isDirectory ()) {
                        file.append ("mimeTypes.rdf");
                        if (file.exists ()) {
                            /* mimeTypes.rdf を発見した */
                            arUnMHTContentHandlerModule._getTypes (file, types);
                        }
                    }
                }
                catch (e) {
                }
            }
        }
        catch (e) {
        }
        
        for (var i = 0; i < types.length; i ++) {
            compMgr.registerFactoryLocation (this.CID,
                                             this.CNAME,
                                             prefix + types [i],
                                             fileSpec, location, type);
        }
    },
    
    /**
     * mimeTypes.rdf から mht の MIME Type を取得する
     *
     * @param  nsIFile file
     *         mimeTypes.rdf
     * @param  Array types
     *         既知の MIME Type
     *           [String Mime Type, ...]
     */
    _getTypes : function (file, types) {
        var rdf
        = Components.classes ["@mozilla.org/rdf/rdf-service;1"]
        .getService (nsIRDFService);
        
        var spec
        = Components.classes ["@mozilla.org/network/io-service;1"]
        .getService (nsIIOService)
        .getProtocolHandler ("file")
        .QueryInterface (nsIFileProtocolHandler)
        .getURLSpecFromFile (file);
        
        var mimeDS = rdf.GetDataSourceBlocking (spec);
        var container
        = Components.classes ["@mozilla.org/rdf/container;1"]
        .createInstance (nsIRDFContainer);
        
        container.Init (mimeDS,
                        rdf.GetResource ("urn:mimetypes:root"));
        var elements = container.GetElements ();
        while (elements.hasMoreElements ()) {
            var fileType = elements.getNext ();
            if (!(fileType instanceof nsIRDFResource)) {
                break;
            }
            
            var ext = "";
            var property = rdf.GetResource
                ("http://home.netscape.com/NC-rdf#fileExtensions");
            var target = rdf.GetLiteral ("mht");
            
            if (mimeDS.HasAssertion (fileType, property, target, true)) {
                /* 拡張子 mht の項目を発見した */
                var userDefined = "";
                target
                    = rdf.GetResource
                    ("http://home.netscape.com/NC-rdf#value");
                value = mimeDS.GetTarget (fileType, target, true);
                if (value) {
                    userDefined
                        = value.QueryInterface (nsIRDFLiteral).Value;
                }
                
                if (userDefined) {
                    var exists = false;
                    
                    for (var i = 0; i < types.length; i ++) {
                        if (userDefined == types [i]) {
                            exists = true;
                            break;
                        }
                    }
                    
                    if (!exists) {
                        types.push (userDefined);
                    }
                }
            }
        }
    },
    
    /**
     * 登録解除処理
     *   nsIModule.unregisterSelf
     *
     * @param  nsIComponentManager compMgr
     * @param  nsIFile fileSpec
     *         モジュールのファイル
     *           reference と他のソースコード中で並びが違う
     * @param  String location
     *         不明
     *           reference と他のソースコード中で並びが違う
     */
    unregisterSelf : function (compMgr, fileSpec, location) {
        try {
            var id
            = Components.classes ["@mozilla.org/xre/app-info;1"]
            .getService (nsIXULAppInfo).ID;
            if (id == "{3550f703-e582-4d05-9a08-453d09bdfdc6}") {
                /* Thunderbird */
            
                return;
            }
        } catch (e) {}
        
        compMgr = compMgr.QueryInterface (nsIComponentRegistrar);
        compMgr.unregisterFactoryLocation (this.CID, fileSpec);
        
        
        
        var catman
        = Components.classes ["@mozilla.org/categorymanager;1"]
        .getService (nsICategoryManager);
        catman.deleteCategoryEntry ("command-line-handler",
                                    this.CONTRACTID, true);
    },
    
    /**
     * ファクトリーオブジェクトを取得する
     *   nsIModule.getClassObject
     *
     * @param  nsIComponentManager compMgr
     * @param  nsCIDRef cid
     *         取得対象のクラス ID
     * @param  nsIIDRef iid
     *         取得対象のインターフェース ID
     * @throws Components.results.NS_ERROR_NOT_IMPLEMENTED
     *         Components.results.NS_ERROR_NO_INTERFACE
     * @return arUnMHTContentHandlerFactory
     *         本体のファクトリー
     */
    getClassObject : function (compMgr, cid, iid) {
        if (cid.equals (this.CID)) {
            return arUnMHTContentHandlerFactory;
        }
        
        if (!iid.equals (nsIFactory)) {
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
    
    /**
     * 終了できるかどうか
     *   nsIModule.canUnload
     *
     * @param  nsIComponentManager compMgr
     * @return Boolean
     *         終了できるかどうか
     */
    canUnload : function (compMgr) {
        return true;
    }
};

/**
 * モジュールを取得する
 * @param  nsIComponentManager compMgr
 * @param  nsIFile fileSpec
 *         モジュールのファイル
 * @return arUnMHTContentHandlerModule
 *         モジュール
 */
function NSGetModule (compMgr, fileSpec) {
    return arUnMHTContentHandlerModule;
}
