/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* Unicode をエスケープ解除できない場合に unescape を定義しなおす */
if (unescape ("%u3042") == "%u3042") {
    unescape = function (text) {
        var converter
        = Components.classes ["@mozilla.org/intl/scriptableunicodeconverter"]
        .getService (Components.interfaces.nsIScriptableUnicodeConverter);
        converter.charset = "utf-8";
        
        text = text
        .replace (/((%[0-9A-Fa-f][0-9A-Fa-f])+)|%u([0-9A-Fa-f][0-9A-Fa-f][0-9A-Fa-f][0-9A-Fa-f])/g,
                  function (match, part1, part2, part3) {
                      if (part1) {
                          var t = part1
                          .replace (/%([0-9A-Fa-f][0-9A-Fa-f])/g,
                                    function (submatch, subpart1) {
                                        return String
                                        .fromCharCode (parseInt ("0x"
                                                                 + subpart1));
                                    });
                          try {
                              t = converter.ConvertToUnicode (t);
                          }
                          catch (e) {
                          }
                          return t;
                      }
                      else {
                          return String
                          .fromCharCode (parseInt ("0x" + part3));
                      }
                  });
        return text;
    }
}

/**
 * 設定管理
 **/
var UnMHTOptions = {
    /**
     * 初期化処理
     */
    init : function () {
        try {
            var id
            = Components.classes ["@mozilla.org/xre/app-info;1"]
            .getService (Components.interfaces.nsIXULAppInfo).ID;
            if (id == "{3550f703-e582-4d05-9a08-453d09bdfdc6}") {
                /* Thunderbird */
                UnMHTOptions.isThunderbird = true;
            }
        } catch (e) {}
        
        if (UnMHTOptions.isThunderbird) {
            document.getElementById ("tab").selectedTab
            = document.getElementById ("tab_other_tab");
            
            document.getElementById ("tab_general_tab").hidden = true;
            document.getElementById ("tab_filter_tab").hidden = true;
            document.getElementById ("tab_save_tab").hidden = true;
            document.getElementById ("toolbar_group").hidden = true;
            document.getElementById ("all").hidden = true;
        }
        
        UnMHTOptions.loadPrefs (null);
    },
    
    /**
     * 適用
     */
    apply : function () {
        UnMHTOptions.savePrefs (null);
    },
    
    /**
     * 終了処理
     *
     * @return Boolean
     *         ダイアログを閉じるか
     */
    term : function () {
        if (UnMHTOptions.checkFocus ()) {
            return UnMHTOptions.savePrefs (null);
        }
        else {
            return false;
        }
    },
    
    /**
     * 設定を読み込む
     * 設定が無ければ既定値を書き込む
     *
     * @param Object map
     *        設定のマップ
     *        prefs.js から読む場合は null
     *        <String 設定名, String 設定値>
     * @param  String type
     *         設定の種類
     *          "bool"
     *          "char"
     *          "int"
     * @param  String name
     *         設定名
     * @param  Boolean/String/Number value
     *         既定値
     * @return Boolean/String/Number
     *         取得した値
     *         設定が無ければ既定値
     */
    initPref : function (map, type, name, value) {
        if (map == null) {
            var prefBranch;
            
            if (Components.interfaces.nsIPrefBranch2) {
                prefBranch
                    = Components.classes ["@mozilla.org/preferences-service;1"]
                    .getService (Components.interfaces.nsIPrefBranch2);
            }
            else {
                prefBranch
                = Components.classes ["@mozilla.org/preferences-service;1"]
                .getService (Components.interfaces.nsIPrefBranch);
            }
            
            if (prefBranch.prefHasUserValue (name)) {
                ; /* switch のインデント用 */
                switch (type) {
                    case "bool":
                        value = prefBranch.getBoolPref (name);
                        break;
                    case "char":
                        value = prefBranch.getCharPref (name);
                        break;
                    case "int":
                        value = prefBranch.getIntPref (name);
                        break;
                }
            }
            else {
                ; /* switch のインデント用 */
                switch (type) {
                    case "bool":
                        prefBranch.setBoolPref (name, value);
                        break;
                    case "char":
                        prefBranch.setCharPref (name, value);
                        break;
                    case "int":
                        prefBranch.setIntPref (name, value);
                        break;
                }
            }
        }
        else {
            for (var name2 in map) {
                if (name2 == name) {
                    ; /* switch のインデント用 */
                    switch (type) {
                        case "bool":
                            value = (map [name2] == "true");
                            break;
                        case "char":
                            value = map [name2];
                            break;
                        case "int":
                            value = parseInt (map [name2]);
                            break;
                    }
                    break;
                }
            }
        }
        
        return value;
    },
    
    /**
     * 設定を読み込む
     *
     * @param Object map
     *        設定のマップ
     *        prefs.js から読む場合は null
     *        <String 設定名, String 設定値>
     */
    loadPrefs : function (map) {
        if (map == null) {
            document.getElementById ("version").value
            += UnMHTVersion;
            
            var button;
            
            button = document
            .getElementById ("unmht_preferences_dialog")
            .getButton ("extra1");
            button.addEventListener
            ("command",
             function () {
                UnMHTOptions.apply ();
            }, true);
            
            button = document
            .getElementById ("unmht_preferences_dialog")
            .getButton ("accept");
            button.addEventListener
            ("mousedown",
             function () {
                arguments [0].explicitOriginalTarget.focus ();
            }, true);
        }
        
        document.getElementById ("all").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.all", true);
        
        document.getElementById ("memorycache").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.memorycache", true);
        if (Components.classes ["@mozilla.org/pipe;1"] == undefined) {
            document.getElementById ("memorycache").checked = false;
            document.getElementById ("memorycache").disabled = true;
        }
        document.getElementById ("nolimit").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.nolimit", true);
        document.getElementById ("fast_extract").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.fast_extract", false);
        
        document.getElementById ("cache_expire_enabled").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.cache_expire.enabled",
                                 true);
        document.getElementById ("cache_expire").value
        = UnMHTOptions.initPref (map, "int",  "unmht.cache_expire", 5);
        
        document.getElementById ("skip_ppt_warning").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.skip_ppt_warning",
                                 false);

        document.getElementById ("open_multiple").value
        = UnMHTOptions.initPref (map, "char", "unmht.open_multiple",
                                 "tabs");
        
        document.getElementById ("filter").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.filter", false);
        
        var value
        = UnMHTOptions
        .initPref (map, "char", "unmht.filter.list", "");
        
        var listbox = document.getElementById ("filter_list");
        var node = listbox.firstChild;
        while (node) {
            var nextNode = node.nextSibling;
            if (node.nodeName == "listitem") {
                listbox.removeChild (node);
            }
            node = nextNode;
        }
        if (value != "") {
            /* 値を解析するだけなので代入はしない */
            value.replace
            (/([^&,]*)&([^&,]*)&([^&,]*),?/g,
             function (matched, type, pattern, r) {
                if  (unescape (r) == "o"
                     || unescape (r) == "x") {
                    UnMHTOptions
                    .addFilterItem (unescape (type),
                                    unescape (pattern),
                                    unescape (r),
                                    null);
                }
            });
        }
        UnMHTOptions.checkFilter ();
        
        document.getElementById ("save_filename").value
        = UnMHTOptions.initPref (map, "char", "unmht.save.filename",
                                 "leafname");
        UnMHTOptions.selectItem (document.getElementById
                                 ("save_filename"));
        value
        = UnMHTOptions.initPref (map, "char", "unmht.save.filename.custom",
                                 "%FILENAME%");
        document.getElementById ("save_filename_custom").value
        = unescape (value);
        value
        = UnMHTOptions.initPref (map, "char", "unmht.save.filename.custom.am",
                                 "a.m.");
        document.getElementById ("save_filename_custom_am").value
        = unescape (value);
        value
        = UnMHTOptions.initPref (map, "char", "unmht.save.filename.custom.pm",
                                 "p.m.");
        document.getElementById ("save_filename_custom_pm").value
        = unescape (value);
        UnMHTOptions.changeSaveFilename ();
        UnMHTOptions.changeFilename ();
        
        document.getElementById ("save_contextmenu").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.contextmenu", true);

        document.getElementById ("save_shortcut").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.shortcut", false);
        value
        = UnMHTOptions
        .initPref (map, "char", "unmht.save.shortcut.keycode", "VK_S");
        document.getElementById ("save_shortcut_keycode").value
        = unescape (value);
        document.getElementById ("save_shortcut_modifiers_alt").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.shortcut.modifiers.alt",
                   true);
        document.getElementById ("save_shortcut_modifiers_ctrl").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.shortcut.modifiers.ctrl",
                   true);
        document.getElementById ("save_shortcut_modifiers_meta").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.shortcut.modifiers.meta",
                   false);
        document.getElementById ("save_shortcut_modifiers_shift").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.shortcut.modifiers.shift",
                   false);
        UnMHTOptions.checkSaveShortcut ();
        
        document.getElementById ("save_inserturl").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.inserturl", false);
        
        document.getElementById ("save_insertlink").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.insertlink", false);
        
        document.getElementById ("save_show_errors").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.show_errors", true);
        
        document.getElementById ("save_skip_alert_on_deep").checked
        = UnMHTOptions
        .initPref (map, "bool", "unmht.save.skip_alert_on_deep", false);
        value
        = UnMHTOptions
        .initPref (map, "char", "unmht.save.skip_alert_on_deep.default",
                   "save");
        document.getElementById ("save_skip_alert_on_deep_default").value
        = unescape (value);
        UnMHTOptions.checkSkipAlertOnDeep ();
        
        document.getElementById ("statusbar_preferences").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.statusbar.preferences",
                                 true);
        document.getElementById ("toolbar_preferences").checked
        = UnMHTOptions.initPref (map, "bool", "unmht.toolbar.preferences",
                                 false);
        
        sizeToContent ();
    },
    
    /**
     * 設定を保存する
     *
     * @param  nsIFileOutputStream fstream
     *         保存先のファイル
     *         prefs.js に保存する場合は null
     * @param  String type
     *         設定の種類
     *          "bool"
     *          "char"
     *          "int"
     * @param  String name
     *         設定名
     * @param  Boolean/String/Number value
     *         既定値
     */
    setPref : function (fstream, type, name, value) {
        if (fstream == null) {
            var prefBranch;
            
            if (Components.interfaces.nsIPrefBranch2) {
                prefBranch
                    = Components.classes ["@mozilla.org/preferences-service;1"]
                    .getService (Components.interfaces.nsIPrefBranch2);
            }
            else {
                prefBranch
                = Components.classes ["@mozilla.org/preferences-service;1"]
                .getService (Components.interfaces.nsIPrefBranch);
            }
            
            ; /* switch のインデント用 */
            switch (type) {
                case "bool":
                    prefBranch.setBoolPref (name, value);
                    break;
                case "char":
                    prefBranch.setCharPref (name, value);
                    break;
                case "int":
                    prefBranch.setIntPref (name, value);
                    break;
            }
        }
        else {
            var line = escape (name) + "," + escape (value) + "\r\n";
            fstream.write (line, line.length);
        }
    },
    
    /**
     * 設定を保存する
     *
     * @param  nsIFileOutputStream fstream
     *         保存先のファイル
     *         prefs.js に保存する場合は null
     * @return Boolean
     *         成功フラグ
     */
    savePrefs : function (fstream) {
        var value;
        
        if (fstream) {
            UnMHTOptions
                .setPref (fstream, "char", "unmht.version",
                          UnMHTVersion);
        }
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.all",
                  document.getElementById ("all").checked);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.memorycache",
                  document.getElementById ("memorycache").checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.nolimit",
                  document.getElementById ("nolimit").checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.fast_extract",
                  document.getElementById ("fast_extract").checked);
        value   = document.getElementById ("cache_expire").value;
        UnMHTOptions
        .setPref (fstream, "int", "unmht.cache_expire", 
                  parseInt (value));
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.cache_expire.enabled",
                  document.getElementById ("cache_expire_enabled").checked);

        UnMHTOptions
        .setPref (fstream, "bool", "unmht.skip_ppt_warning",
                  document.getElementById ("skip_ppt_warning").checked);

        UnMHTOptions
        .setPref (fstream, "char", "unmht.open_multiple",
                  document.getElementById ("open_multiple").value);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.filter",
                  document.getElementById ("filter").checked);
        
        var listbox = document.getElementById ("filter_list");
        var tmp = "";
        var node = listbox.firstChild;
        while (node) {
            var nextNode = node.nextSibling;
            if (node.nodeName == "listitem") {
                var value = UnMHTOptions.getFilterItem (node);
                
                if (tmp != "") {
                    tmp += ",";
                }
                
                tmp
                    += escape (value [0])
                    + "&" + escape (value [1])
                    + "&" + escape (value [2]);
            }
            node = nextNode;
        }
        UnMHTOptions
        .setPref (fstream, "char", "unmht.filter.list",
                  tmp);
        
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.filename",
                  document.getElementById ("save_filename").value);
        value = document.getElementById ("save_filename_custom").value;
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.filename.custom",
                  escape (value));
        value = document.getElementById ("save_filename_custom_am").value;
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.filename.custom.am",
                  escape (value));
        value = document.getElementById ("save_filename_custom_pm").value;
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.filename.custom.pm",
                  escape (value));

        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.contextmenu",
                  document.getElementById ("save_contextmenu").checked);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.shortcut",
                  document.getElementById ("save_shortcut").checked);
        value   = document.getElementById ("save_shortcut_keycode").value;
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.shortcut.keycode",
                  escape (value));
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.shortcut.modifiers.alt",
                  document.getElementById ("save_shortcut_modifiers_alt")
                  .checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.shortcut.modifiers.ctrl",
                  document.getElementById ("save_shortcut_modifiers_ctrl")
                  .checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.shortcut.modifiers.meta",
                  document.getElementById ("save_shortcut_modifiers_meta")
                  .checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.shortcut.modifiers.shift",
                  document.getElementById ("save_shortcut_modifiers_shift")
                  .checked);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.inserturl",
                  document.getElementById ("save_inserturl").checked);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.insertlink",
                  document.getElementById ("save_insertlink").checked);
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.show_errors",
                  document.getElementById ("save_show_errors").checked);

        UnMHTOptions
        .setPref (fstream, "bool", "unmht.save.skip_alert_on_deep",
                  document.getElementById ("save_skip_alert_on_deep").checked);
        value
        = document.getElementById ("save_skip_alert_on_deep_default").value
        UnMHTOptions
        .setPref (fstream, "char", "unmht.save.skip_alert_on_deep.default",
                  escape (value));
        
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.statusbar.preferences",
                  document.getElementById ("statusbar_preferences").checked);
        UnMHTOptions
        .setPref (fstream, "bool", "unmht.toolbar.preferences",
                  document.getElementById ("toolbar_preferences").checked);
        
        UnMHTOptions
        .setPref (fstream, "char", "unmht.savepref",
                  new Date ().getTime ());
        
        if (fstream == null) {
            var prefBranch;
            
            var prefService
                = Components.classes ["@mozilla.org/preferences-service;1"].
                getService (Components.interfaces.nsIPrefService);
            
            prefService.savePrefFile (null);
        }
        
        return true;
    },
    
    /**
     * ラヂオボタンの選択肢を表示に反映する
     * (古いバージョンだと value の変更だけでは反映されない)
     *
     */
    selectItem : function (node) {
        var value = node.value;
        
        var nodes = node.getElementsByTagName ("radio");
        for (var i = 0; i < nodes.length; i ++) {
            if (nodes [i].value == value) {
                node.selectedItem = nodes [i];
            }
        }
    },
    
    /**
     * フィルタを追加するボタンのイベント
     *
     * @param  Boolean modify
     *         修正するか追加するか
     *           true: 修正
     *           false: 追加
     */
    onAddFilter : function (modify) {
        var type = document.getElementById ("filter_type").value;
        var pattern = document.getElementById ("filter_pattern").value;
        var r;
        document.getElementById ("filter_illegal").value = "";
        
        if (document.getElementById ("filter_regexp").checked) {
            r = "o";
            try {
                var tmp = "test";
                tmp.search (pattern);
            }
            catch (e) {
                document.getElementById ("filter_illegal").value
                    = document.getElementById ("filter_illegal_regexp").value;
                
                return;
            }
        }
        else {
            r = "x";
        }
        
        var listbox = document.getElementById ("filter_list");
        var exist = false;
        
        if (type && pattern) {
            var nextNode = null;
            var node = null;
            if (modify) {
                var node = listbox.firstChild;
                while (node) {
                    if (node.nodeName == "listitem") {
                        if (node.selected) {
                            break;
                        }
                    }
                    node = node.nextSibling;
                }
            }
            
            var node2 = listbox.firstChild;
            while (node2) {
                var nextNode2 = node2.nextSibling;
                
                var value = UnMHTOptions.getFilterItem (node2);
                if (type == value [0] && pattern == value [1]
                    && r == value [2]
                    && node2 != node) {
                    exist = true;
                    break;
                }
                node2 = nextNode2;
            }
            
            if (!exist) {
                UnMHTOptions.addFilterItem (type, pattern, r, node);
                
                if (node == null) {
                    document.getElementById ("filter_pattern").value
                        = "";
                }
            }
            else {
                document.getElementById ("filter_illegal").value
                    = document.getElementById ("filter_illegal_dup").value;
            }
        }
        else {
            if (pattern == "") {
                document.getElementById ("filter_illegal").value
                = document.getElementById ("filter_illegal_nullpat").value;
            }
            else {
                document.getElementById ("filter_illegal").value
                = document.getElementById ("filter_illegal_nulllink").value;
            }
        }
    },
    
    /**
     * リストの選択項目を上にずらすボタンのイベント
     *
     * @param  String id
     *         リストの id
     */
    onMoveupList : function (id) {
        var listbox = document.getElementById (id);
        
        var node = listbox.firstChild;
        var nextNode = null;
        var prevNode = null;
        var node1 = null;
        var node2 = null;
        var tmp = "";
        while (node) {
            nextNode = node.nextSibling;
            if (node.nodeName == "listitem") {
                if (node.selected) {
                    if (prevNode) {
                        node1 = node.firstChild;
                        node2 = prevNode.firstChild;
                        while (node1 && node2) {
                            tmp = node1.getAttribute ("value");
                            node1.setAttribute ("value",
                                                node2.getAttribute ("value"));
                            node2.setAttribute ("value", tmp);
                            
                            if (node1.getAttribute ("class")
                                == "listcell-iconic") {
                                tmp = node1.getAttribute ("image");
                                node1.setAttribute ("image",
                                                    node2
                                                    .getAttribute ("image"));
                                node2.setAttribute ("image", tmp);
                            }
                            else {
                                tmp = node1.getAttribute ("label");
                                node1.setAttribute ("label",
                                                    node2
                                                    .getAttribute ("label"));
                                node2.setAttribute ("label", tmp);
                            }
                            
                            node1 = node1.nextSibling;
                            node2 = node2.nextSibling;
                        }
                        listbox.toggleItemSelection (node);
                        listbox.toggleItemSelection (prevNode);
                        prevNode = prevNode;
                    }
                }
                else {
                    prevNode = node;
                }
            }
            node = nextNode;
        }
    },
    
    /**
     * リストの選択項目を下にずらすボタンのイベント
     *
     * @param  String id
     *         リストの id
     */
    onMovedownList : function (id) {
        var listbox = document.getElementById (id);
        
        var node = listbox.lastChild;
        var nextNode = null;
        var prevNode = null;
        var node1 = null;
        var node2 = null;
        var tmp = "";
        while (node) {
            prevNode = node.previousSibling;
            if (node.nodeName == "listitem") {
                if (node.selected) {
                    if (nextNode) {
                        node1 = node.firstChild;
                        node2 = nextNode.firstChild;
                        while (node1 && node2) {
                            tmp = node1.getAttribute ("value");
                            node1.setAttribute ("value",
                                                node2.getAttribute ("value"));
                            node2.setAttribute ("value", tmp);
                            
                            if (node1.getAttribute ("class")
                                == "listcell-iconic") {
                                tmp = node1.getAttribute ("image");
                                node1.setAttribute ("image",
                                                    node2
                                                    .getAttribute ("image"));
                                node2.setAttribute ("image", tmp);
                            }
                            else {
                                tmp = node1.getAttribute ("label");
                                node1.setAttribute ("label",
                                                    node2
                                                    .getAttribute ("label"));
                                node2.setAttribute ("label", tmp);
                            }
                            
                            node1 = node1.nextSibling;
                            node2 = node2.nextSibling;
                        }
                        listbox.toggleItemSelection (node);
                        listbox.toggleItemSelection (nextNode);
                        nextNode = nextNode;
                    }
                }
                else {
                    nextNode = node;
                }
            }
            node = prevNode;
        }
    },
    
    /**
     * フィルタを削除するボタンのイベント
     */
    onDeleteFilter : function () {
        var listbox = document.getElementById ("filter_list");
        
        var node = listbox.firstChild;
        while (node) {
            var nextNode = node.nextSibling;
            if (node.selected) {
                listbox.removeChild (node);
            }
            node = nextNode;
        }
    },
    
    /**
     * フィルタを選択したイベント
     */
    selectFilter : function () {
        var listbox = document.getElementById ("filter_list");
        
        if (listbox.selectedIndex != -1) {
            var value = UnMHTOptions
                .getFilterItem (listbox.selectedItem);
            
            document.getElementById ("filter_type").value
                = value [0];
            document.getElementById ("filter_pattern").value
                = value [1];
            document.getElementById ("filter_regexp").checked
                = value [2] == "o";
            
            document.getElementById ("filter_modify").disabled
                = !document.getElementById ("filter").checked
                || (listbox.selectedCount > 1);
        }
    },
    
    /**
     * フィルタを取得する
     *
     * @param  Listitem listitem 
     *         対象の項目
     * @return Array
     *         フィルタ
     *           [String パターン,
     *            String 正規表現ならば "o" そうでなければ "x", String リンク先]
     */
    getFilterItem : function (listitem) {
        var listcell1 = listitem.firstChild;
        var listcell2 = listcell1.nextSibling;
        var listcell3 = listcell2.nextSibling;
        return new Array (listcell1.getAttribute ("value"),
                          listcell2.getAttribute ("value"),
                          listcell3.getAttribute ("value"));
    },
    
    /**
     * フィルタを追加する
     *
     * @param  String type
     *         フィルタの種類
     * @param  String pattern
     *         パターン
     * @param  String r
     *         正規表現フラグ
     *             "o"
     *             "x"
     * @param  XULElement listitem
     *         置き換えるノード
     *         追加する場合は null
     */
    addFilterItem : function (type, pattern, r, listitem) {
        var listbox = document.getElementById ("filter_list");
        var listcell;
        var append = false;
        
        if (listitem == null) {
            append = true;
            listitem = document.createElement ("listitem");
            
            listcell = document.createElement ("listcell");
            listitem.appendChild (listcell);
            listcell = document.createElement ("listcell");
            listitem.appendChild (listcell);
            listcell = document.createElement ("listcell");
            listitem.appendChild (listcell);
            
            listitem.addEventListener
                ("mousedown",
                 function () {
                    UnMHTOptions.onMouseDownFilter (arguments [0]);
                }, false);
        }
        
        var listcell = listitem.firstChild;
        listcell.setAttribute ("label",
                               document.getElementById
                               ("filter_type_" + type).value);
        listcell.setAttribute ("value", type);
        
        listcell = listcell.nextSibling;
        listcell.setAttribute ("label", pattern);
        listcell.setAttribute ("value", pattern);
        
        listcell = listcell.nextSibling;
        listcell.setAttribute ("value", r);
        listcell.setAttribute ("class", "listcell-iconic");
        if (r == "o") {
            listcell.setAttribute ("image",
                                   "chrome://unmht/content/check_o.png");
        }
        else {
            listcell.setAttribute ("image",
                                   "chrome://unmht/content/check_x.png");
        }
        
        if (append) {
            listbox.appendChild (listitem);
        }
    },
    
    /**
     * フィルタの項目を押したイベント
     *
     * @param  Event event
     *         対象のイベント
     */
    onMouseDownFilter : function (event) {
        var listitem = event.target;
        if (listitem.nodeName.toLowerCase () != "listitem") {
            return;
        }
        
        var listcell1 = listitem.firstChild;
        var listcell2 = listcell1.nextSibling;
        var listcell3 = listcell2.nextSibling;
        
        var value = UnMHTOptions.getFilterItem (listitem);
        
        if (event.clientX
            < listcell1.boxObject.x + listcell1.boxObject.width) {
            return;
        }
        else if (event.clientX
                 < listcell2.boxObject.x + listcell2.boxObject.width) {
            return;
        }
        else if (event.clientX
                 < listcell3.boxObject.x + listcell3.boxObject.width) {
            if (event.clientX < listcell3.boxObject.x + 16) {
                value [2] = (value [2] == "o") ? "x" : "o";
            }
        }
        
        document.getElementById ("filter_type").value
        = value [0];
        document.getElementById ("filter_pattern").value
        = value [1];
        document.getElementById ("filter_regexp").checked
        = value [2] == "o";
        
        UnMHTOptions.addFilterItem (value [0], value [1],
                                    value [2], listitem);
    },
    
    /* 親のチェックボックスの状態で子の操作不可を設定する - ここから - */
    
    checkFilter : function () {
        document.getElementById ("filter_type").disabled
        = document.getElementById ("filter_type_label").disabled
        = document.getElementById ("filter_pattern_label").disabled
        = document.getElementById ("filter_pattern").disabled
        = document.getElementById ("filter_regexp").disabled
        = document.getElementById ("filter_list").disabled
        = document.getElementById ("filter_add").disabled
        = document.getElementById ("filter_moveup").disabled
        = document.getElementById ("filter_movedown").disabled
        = document.getElementById ("filter_delete").disabled
        = !document.getElementById ("filter").checked;
        
        document.getElementById ("filter_modify").disabled
        = !document.getElementById ("filter").checked
        || (document.getElementById ("filter_list").selectedCount > 1);
    },
    
    /* 親のチェックボックスの状態で子の操作不可を設定する - ここまで - */
    
    /**
     * 設定のインポート
     */
    importPrefs : function () {
        try {
            var filename;
            
            var filePicker
            = Components.classes ["@mozilla.org/filepicker;1"]
            .createInstance (Components.interfaces.nsIFilePicker);
            filePicker.init (window,
                             document.getElementById ("import_title").value,
                             Components.interfaces.nsIFilePicker.modeOpen);
            filePicker.appendFilter ("Text", "*.txt");
            filePicker.defaultString = "unmhtConfig.txt";
            filePicker.appendFilters (Components.interfaces.nsIFilePicker
                                      .filterAll);
            
            var ret = filePicker.show ();
            if (ret == Components.interfaces.nsIFilePicker.returnOK) {
                var file = filePicker.file
                    .QueryInterface (Components.interfaces.nsILocalFile);
                
                var fstream
                    = Components
                    .classes ["@mozilla.org/network/file-input-stream;1"]
                    .createInstance (Components.interfaces.nsIFileInputStream);
                var sstream
                    = Components
                    .classes ["@mozilla.org/scriptableinputstream;1"]
                    .createInstance (Components.interfaces
                                     .nsIScriptableInputStream);
                fstream.init (file, 0x01, 0444, 0);
                sstream.init (fstream);
                var text = sstream.read (-1);
                sstream.close ();
                fstream.close ();
                
                text = text.replace (/\r\n/g, "\n");
                var map = new Object ();
                var ok = false;
                
                /* ファイルを解析するだけなので代入はしない */
                text.replace
                    (/([^\n]+)\n?/g,
                     function (matched, part1) {
                        if (part1.match (/^(unmht\..+),(.+)$/)) {
                            map [unescape (RegExp.$1)] = unescape (RegExp.$2);
                            ok = true;
                        }
                        return "";
                    });
                
                if (ok) {
                    UnMHTOptions.loadPrefs (map);
                    
                    document.getElementById ("import_export_message").value
                        = document.getElementById ("import_done").value;
                }
                else {
                    document.getElementById ("import_export_message").value
                        = document.getElementById ("import_illegal").value;
                }
            }
            else {
                document.getElementById ("import_export_message").value
                = document.getElementById ("import_abort").value;
            }
        }
        catch (e) {
            document.getElementById ("import_export_message").value
            = document.getElementById ("import_fail").value;
        }
    },
    
    /**
     * 設定のエクスポート
     */
    exportPrefs : function () {
        try {
            var filename;
            
            var filePicker
            = Components.classes ["@mozilla.org/filepicker;1"]
            .createInstance (Components.interfaces.nsIFilePicker);
            filePicker.init (window,
                             document.getElementById ("export_title").value,
                             Components.interfaces.nsIFilePicker.modeSave);
            filePicker.appendFilter ("Text", "*.txt");
            filePicker.defaultString = "unmhtConfig.txt";
            filePicker.appendFilters (Components.interfaces.nsIFilePicker
                                      .filterAll);
            
            var ret = filePicker.show ();
            if (ret == Components.interfaces.nsIFilePicker.returnOK ||
                ret == Components.interfaces.nsIFilePicker.returnReplace) {
                var file
                    = filePicker.file
                    .QueryInterface (Components.interfaces.nsILocalFile);
                
                var fstream
                    = Components
                    .classes ["@mozilla.org/network/file-output-stream;1"]
                    .createInstance (Components.interfaces.nsIFileOutputStream);
                fstream.init (file, 0x02 | 0x08 | 0x20, 0664, 0);
                
                UnMHTOptions.savePrefs (fstream);
                
                fstream.close ();
                
                document.getElementById ("import_export_message").value
                    = document.getElementById ("export_done").value;
            }
            else {
                document.getElementById ("import_export_message").value
                = document.getElementById ("export_abort").value;
            }
        }
        catch (e) {
            document.getElementById ("import_export_message").value
            = document.getElementById ("export_fail").value;
        }
    },
    
    /**
     * 設定の初期化
     */
    initPrefs : function () {
        var map = new Object ();
        UnMHTOptions.loadPrefs (map);
    },
    
    /**
     * フォーカスされている要素を調べる
     * テキストボックス内でエンターを押してもダイアログが閉じないようにする
     *
     * @return Boolean
     *         1行のテキストボックスにフォーカスが無ければ true
     */
    checkFocus : function () {
        var node = document.commandDispatcher.focusedElement;
        
        if  (node
             && node.nodeName == "html:input") {
            return false;
        }
        
        return true;
    },
    
    /**
     * サイトを開く
     */
    openWebsite : function () {
        var mediator
        = Components.classes ["@mozilla.org/appshell/window-mediator;1"]
        .getService (Components.interfaces.nsIWindowMediator);
        var chromeWindow = mediator.getMostRecentWindow ("navigator:browser");
        if (chromeWindow) {
            chromeWindow.UnMHTUIManager.openWebsite ();
        }
        else {
            var locale = "ja-JP";
            try {
                var prefBranch;
            
                if (Components.interfaces.nsIPrefBranch2) {
                    prefBranch
                        = Components.classes
                        ["@mozilla.org/preferences-service;1"]
                        .getService (Components.interfaces.nsIPrefBranch2);
                }
                else {
                    prefBranch
                        = Components.classes
                        ["@mozilla.org/preferences-service;1"]
                        .getService (Components.interfaces.nsIPrefBranch);
                }
                var name = "general.useragent.locale";
                if (prefBranch.prefHasUserValue (name)) {
                    locale = prefBranch.getCharPref (name);
                }
            }
            catch (e) {
            }
            if (locale == "ja-JP") {
                window.open ("http://www.unmht.org/unmht/");
            }
            else {
                window.open ("http://www.unmht.org/unmht/en_index.html");
            }
        }
    },
    
    /**
     * MIME Type を修正する
     */
    checkMIME : function () {
        var exists = false;
        document.getElementById ("checkmime_message").value = "";
        try {
            var file
                = Components.classes ["@mozilla.org/file/directory_service;1"]
                .getService (Components.interfaces.nsIProperties)
                .get ("UMimTyp", Components.interfaces.nsIFile);
            
            if (UnMHTOptions.modifyMIME (file)) {
                exists = true;
            }
        }
        catch (e) {
        }
        try {
            var profdir
            = Components.classes ["@mozilla.org/file/directory_service;1"]
            .getService (Components.interfaces.nsIProperties)
            .get ("DefProfRt", Components.interfaces.nsIFile);
            var entries = profdir.directoryEntries;
            while (entries.hasMoreElements ()) {
                file = entries.getNext ();
                try {
                    file = file.QueryInterface (Components.interfaces.nsIFile);
                    if (file.isDirectory ()) {
                        file.append ("mimeTypes.rdf");
                        if (file.exists ()) {
                            /* mimeTypes.rdf を発見した */
                        
                            if (UnMHTOptions.modifyMIME (file)) {
                                exists = true;
                            }
                        }
                    }
                }
                catch (e) {
                }
            }
        }
        catch (e) {
        }
        
        if (exists) {
            document.getElementById ("checkmime_message").value
            = document.getElementById ("checkmime_ok").value;
        }
        else {
            document.getElementById ("checkmime_message").value
            = document.getElementById ("checkmime_no").value;
        }
    },
    
    /**
     * mimeTypes.rdf を修正する
     *
     * @param  nsIFile file
     *         mimeTypes.rdf
     */
    modifyMIME : function (file) {
        var rdf
        = Components.classes ["@mozilla.org/rdf/rdf-service;1"]
        .getService (Components.interfaces.nsIRDFService);
            
        var result = false;
        
        var spec
        = Components.classes ["@mozilla.org/network/io-service;1"]
        .getService (Components.interfaces.nsIIOService)
        .getProtocolHandler ("file")
        .QueryInterface (Components.interfaces.nsIFileProtocolHandler)
        .getURLSpecFromFile (file);
        
        var mimeDS = rdf.GetDataSource (spec);
        
        var source = rdf.GetUnicodeResource
        ("urn:mimetype:application/x-maf");
        var property = rdf.GetUnicodeResource
        ("http://home.netscape.com/NC-rdf#fileExtensions");
        var target = rdf.GetLiteral ("mht");
        if (mimeDS.HasAssertion (source, property, target, true)) {
            mimeDS.Unassert (source, property, target);
            result = true;
        }
        
        source = rdf.GetUnicodeResource
        ("urn:mimetype:application/maf");
        property = rdf.GetUnicodeResource
        ("http://home.netscape.com/NC-rdf#fileExtensions");
        target = rdf.GetLiteral ("mht");
        if (mimeDS.HasAssertion (source, property, target, true)) {
            mimeDS.Unassert (source, property, target);
            result = true;
        }
        
        mimeDS.QueryInterface (Components.interfaces.nsIRDFRemoteDataSource)
        .Flush ();
        
        var compregFile = file.parent;
        compregFile.append ("compreg.dat");
        if (compregFile.exists ()) {
            compregFile.remove (true);
        }
        
        return result;
    },
    
    /**
     * 保存のキー変更イベント
     *
     * @param  Event event
     *         対象のイベント
     */
    onSaveShortcutKeyDown : function (event) {
        var keycode = "";
        for (var attribute in Components.interfaces.nsIDOMKeyEvent) {
            if (attribute.match (/^DOM_(VK_.+)$/)) {
                var code = RegExp.$1;
                if (event.keyCode
                    == Components.interfaces.nsIDOMKeyEvent [attribute]) {
                    keycode = code;
                }
            }
        }
        if (keycode == "VK_SHIFT"
            || keycode == "VK_CONTROL"
            || keycode == "VK_ALT"
            || keycode == "VK_META") {
            return;
        }
        
        if (keycode != "") {
            document.getElementById ("save_shortcut_keycode").value
            = keycode;
            UnMHTOptions.selectItem (document.getElementById
                                     ("save_shortcut_keycode"));
            
            document.getElementById ("save_shortcut_modifiers_alt").checked
            = event.altKey;
            document.getElementById ("save_shortcut_modifiers_ctrl").checked
            = event.ctrlKey;
            document.getElementById ("save_shortcut_modifiers_meta").checked
            = event.metaKey;
            document.getElementById ("save_shortcut_modifiers_shift").checked
            = event.shiftKey;
        }
        
        event.preventDefault ();
        event.stopPropagation ();
    },
    
    checkSaveShortcut : function () {
        document.getElementById ("save_shortcut_input").disabled
        = document.getElementById ("save_shortcut_keycode").disabled
        = document.getElementById ("save_shortcut_keycode_menu").disabled
        = document.getElementById ("save_shortcut_modifiers_alt").disabled
        = document.getElementById ("save_shortcut_modifiers_ctrl").disabled
        = document.getElementById ("save_shortcut_modifiers_meta").disabled
        = document.getElementById ("save_shortcut_modifiers_shift").disabled
        = !document.getElementById ("save_shortcut").checked;
    },
    
    checkSkipAlertOnDeep : function () {
        document.getElementById ("save_skip_alert_on_deep_default").disabled
        = !document.getElementById ("save_skip_alert_on_deep").checked;
    },
    
    insertCustom : function (text) {
        if (document.getElementById ("save_filename").value != "custom") {
            return;
        }
        
        var textbox = document.getElementById ("save_filename_custom");
        var value = textbox.value;
        var pos = textbox.selectionStart
        
        value = value.substr (0, pos) + text + value.substr (pos);
        
        textbox.value = value;
        
        textbox.focus ();
        
        textbox.selectionStart = textbox.selectionEnd = pos + text.length;
        
        UnMHTOptions.changeFilename ();
    },
    
    changeFilename : function () {
        var TITLE = "no title";
        var FILENAME = "index.html";
        var URL = "http://localhost/index.html";
        var pattern
        = document.getElementById ("save_filename_custom").value;
        var am
        = document.getElementById ("save_filename_custom_am").value;
        var pm
        = document.getElementById ("save_filename_custom_pm").value;
        document.getElementById ("save_filename_custom_sample").value
        = UnMHTFormatter.format (pattern, TITLE, FILENAME, URL, am, pm);
    },
    
    changeSaveFilename : function () {
        document.getElementById ("save_filename_custom_label").disabled
        = document.getElementById ("save_filename_custom").disabled
        = document.getElementById ("save_filename_custom_button_YYYY").disabled
        = document.getElementById ("save_filename_custom_button_YY").disabled
        = document.getElementById ("save_filename_custom_button_MM").disabled
        = document.getElementById ("save_filename_custom_button_M").disabled
        = document.getElementById ("save_filename_custom_button_DD").disabled
        = document.getElementById ("save_filename_custom_button_D").disabled
        = document.getElementById ("save_filename_custom_button_HH").disabled
        = document.getElementById ("save_filename_custom_button_H").disabled
        = document.getElementById ("save_filename_custom_button_hh").disabled
        = document.getElementById ("save_filename_custom_button_h").disabled
        = document.getElementById ("save_filename_custom_button_mm").disabled
        = document.getElementById ("save_filename_custom_button_m").disabled
        = document.getElementById ("save_filename_custom_button_ss").disabled
        = document.getElementById ("save_filename_custom_button_s").disabled
        = document.getElementById ("save_filename_custom_button_ampm").disabled
        = document.getElementById ("save_filename_custom_am_label").disabled
        = document.getElementById ("save_filename_custom_am").disabled
        = document.getElementById ("save_filename_custom_pm_label").disabled
        = document.getElementById ("save_filename_custom_pm").disabled
        = document.getElementById ("save_filename_custom_button_TITLE").disabled
        = document.getElementById ("save_filename_custom_button_FILENAME").disabled
        = document.getElementById ("save_filename_custom_button_URL").disabled
        = document.getElementById ("save_filename_custom_sample").disabled
        = document.getElementById ("save_filename_custom_sample_label").disabled
        = (document.getElementById ("save_filename").value != "custom");
    }
};
