/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/**
 * ダウンロード履歴に追加するためのダミーのリスナ
 */
function UnMHTSaveDownloadListener () {
}
UnMHTSaveDownloadListener.prototype = {
    dl : null,    /* nsIDownload  対象のダウンロード */
    fileSize : 0, /* Number  ファイルサイズ */
    
    /**
     * インターフェースの要求
     *   nsISupports.QueryInterface
     *
     * @param  nsIIDRef iid
     *         インターフェース ID
     * @throws Components.results.NS_NOINTERFACE
     * @return nsICacheListener
     *         this
     */
    QueryInterface : function(iid) {
        if (iid.equals(Components.interfaces.nsISupports)
            || iid.equals(Components.interfaces.nsIStreamListener)
            || iid.equals(Components.interfaces.nsICancelable)) {
            return this;
        }
        
        throw Components.results.NS_NOINTERFACE;
    },
    
    /**
     * キャンセルイベント
     *   nsICancelable.cancel
     */
    cancel: function () {
    },
    
    /**
     * リクエスト開始のイベント
     *   nsIRequestObserver.onStartRequest
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     */
    onStartRequest : function (request, context) {
    },
    
    /**
     * データ到着のイベント
     *   nsIStreamListener.onDataAvailable
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     * @param  nsIInputStream inputStream
     *         データを取得するストリーム
     * @param  PRUint32 offset
     *         データの位置
     * @param  PRUint32 count 
     *         データの長さ
     */
    onDataAvailable : function (request, context, inputStream, offset, count) {
    },
    
    /**
     * リクエスト終了のイベント
     *   nsIRequestObserver.onStopRequest
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     * @param  Number statusCode
     *         終了コード
     */
    onStopRequest : function (request, context, statusCode) {
        this.dl.onProgressChange
        (null, request,
         this.fileSize, this.fileSize,
         this.fileSize, this.fileSize);
        this.dl.onStateChange
        (null, request,
         Components.interfaces.nsIWebProgressListener.STATE_STOP
         | Components.interfaces.nsIWebProgressListener.STATE_IS_NETWORK,
         Components.results.NS_OK);
    }
};


/**
 * 保存するファイルの情報
 *   Inherits From: nsIStreamListener, nsIRequestObserver
 *                  nsICacheListener
 *
 * @param  String loc
 *         Content-Location の値
 */
function UnMHTSaveFile (loc) {
    this.location = loc;
    if (UnMHTSave.protocolHandler.isUnMHTURI (this.location)) {
        var param = UnMHTSave.protocolHandler.getUnMHTURIParam (this.location);
    
        var entries
            = Components.classes ["@mozilla.org/appshell/window-mediator;1"]
            .getService (Components.interfaces.nsIWindowMediator)
            .getEnumerator ("navigator:browser");
    
        var original = param.original;
        var refName = param.refName;
        
        while (entries.hasMoreElements ()) {
            var w = entries.getNext ();
        
            if (original in w.UnMHT.extractors) {
                var ex = w.UnMHT.extractors [original];
            
                var srcFile = null;
                
                for (i = 0; i < ex.files.length; i ++) {
                    srcFile = ex.files [i];
                    if ((srcFile == ex.rootFile
                         && refName == "")
                        || refName == srcFile.refName) {
                        this.headerList = srcFile.headerList;
                        this.extractor = ex;
                        break;
                    }
                }
                break;
            }
        }
    }
}
UnMHTSaveFile.prototype = {
    startTime : 0,        /* Number  開始時間 */
    status : 0,           /* Number  取得状況 */
    depth : 0,            /* Number 参照の深さ */
    base : "",            /* String  ベースディレクトリ */
    isCSS : false,        /* Boolean  CSS かどうか
                           *   CSS の import をチェックするかどうか */
    isHTML : false,       /* Boolean  HTML かどうか
                           *   コメント, リンクを入れるかどうか */
    isText : false,       /* Boolean  テキストかどうか */
    notEncode : false,    /* Boolean  エンコードしない */
    location : "",        /* String  Content-Location の値 */
    originalContent : "", /* String  ファイルの中身 */
    encodedIndex : 0,     /* Number  エンコード済みの位置 */
    fileSize : "",        /* String  ファイルサイズ 不明の場合 unknown */
    recvSize : 0,         /* Number  取得したファイルサイズ */
    content : "",         /* String  base 64 エンコードしたファイルの中身 */
    filetype : "",        /* String  Content-Type のうち、MIME タイプ */
    charset : "",         /* String  Content-Type のうち、文字コード */
    channel : null,       /* nsIChannel  ネットワーク取得用チャネル */
    httpChannel : null,   /* nsIChannel  HTTP ヘッダ取得用チャネル */
    importFiles : null,   /* Array  @import しているファイル
                           *   [String  URL, ...] */
    headerList : null,    /* Array  ヘッダ
                           *   [[String フィールド名, String フィールド値], ...] */
    extractor : null,     /* UnMHTExtractor  MHT ファイルの保持者 */
    childFiles : null,    /* Array  含むファイル (multipart の場合)
                           *   [UnMHTSaveFile, ...] */
    
    /**
     * インターフェースの要求
     *   nsISupports.QueryInterface
     *
     * @param  nsIIDRef iid
     *         インターフェース ID
     * @throws Components.results.NS_NOINTERFACE
     * @return nsICacheListener
     *         this
     */
    QueryInterface : function (iid) {
        if (iid.equals (Components.interfaces.nsISupports)
            || iid.equals (Components.interfaces.nsISupportsWeakReference)
            || iid.equals (Components.interfaces.nsIStreamListener)
            || iid.equals (Components.interfaces.nsIRequestObserver)
            || iid.equals (Components.interfaces.nsICacheListener)) {
            return this;
        }
        
        throw Components.results.NS_NOINTERFACE;
    },
    
    /**
     * リクエスト開始のイベント
     *   nsIRequestObserver.onStartRequest
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     */
    onStartRequest : function (request, context) {
        this.originalContent = "";
        this.content = "";
        this.encodedIndex = 0;
        try {
            if (this.channel) {
                this.httpChannel
                    = this.channel.QueryInterface
                    (Components.interfaces.nsIHttpChannel);
            
                if (this.httpChannel.responseStatus >= 400) {
                    this.status = UnMHTSave.STATUS_NG;
                    
                    var NS_BINDING_ABORTED = 0x804b0002;
                    try {
                        this.channel.cancel (NS_BINDING_ABORTED);
                    }
                    catch (e) {
                    }
                    
                    return;
                }
                
                this.fileSize = "";
                try {
                    this.fileSize
                        = this.httpChannel.getResponseHeader ("Content-Length");
                
                }
                catch (e) {
                }
                if (this.fileSize && parseInt (this.fileSize) > 0) {
                    this.fileSize += " bytes";
                }
                else {
                    this.fileSize = "unknown";
                }
            }
        }
        catch (e) {
        }
    },
    
    /**
     * リクエスト終了のイベント
     *   nsIRequestObserver.onStopRequest
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     * @param  Number statusCode
     *         終了コード
     */
    onStopRequest : function (request, context, statusCode) {
        if (this.status == UnMHTSave.STATUS_NG) {
            return;
        }
        if (statusCode != Components.results.NS_OK) {
            this.status = UnMHTSave.STATUS_NG;
            return;
        }
        if (this.filetype == ""
            && this.httpChannel) {
            try {
                var tmp
                = this.httpChannel.getResponseHeader ("Content-Type");
                if (tmp.match (/([^ ;]+\/[^ ;]+)/)) {
                    this.filetype = RegExp.$1;
                }
            }
            catch (e) {
            }
        }
        if (this.filetype == "") {
            if (this.location.match (/\.[sx]?html?(\?.+)?(#.+)?$/i)) {
                this.filetype = "text/html";
            }
            else if (this.location.match (/\.css(\?.+)?$/i)) {
                this.filetype = "text/css";
            }
            else if (this.location.match (/\.mht(\?.+)?$/i)) {
                this.filetype = "message/rfc822";
            }
            else if (this.location.match (/\.js(\?.+)?$/i)) {
                this.filetype = "text/javascript";
            }
            else if (this.location.match (/\.txt(\?.+)?$/i)) {
                this.filetype = "text/plain";
            }
            else if (this.location.match (/\.jpe?g(\?.+)?$/i)) {
                this.filetype = "image/jpeg";
            }
            else if (this.location.match (/\.gif(\?.+)?$/i)) {
                this.filetype = "image/gif";
            }
            else if (this.location.match (/\.png(\?.+)?$/i)) {
                this.filetype = "image/png";
            }
            else if (this.location.match (/\.bmp(\?.+)?$/i)) {
                this.filetype = "image/bmp";
            }
            else if (this.location.match (/\.ico(\?.+)?$/i)) {
                this.filetype = "image/x-icon";
            }
            else if (this.location.match (/\.svg(\?.+)?$/i)) {
                this.filetype = "image/svg+xml";
            }
            else if (this.location.match (/\.swf(\?.+)?$/i)) {
                this.filetype = "application/swf";
            }
            else if (this.location.match (/\.jar(\?.+)?$/i)) {
                this.filetype = "application/java-archive";
            }
            else if (this.location.match (/\.class(\?.+)?$/i)) {
                this.filetype = "application/octet-stream";
            }
            else if (this.location.match (/^about:blank$/)) {
                this.filetype = "text/html";
            }
            else if (this.location.match (/\.([A-Za-z0-9])(\?.*)?$/i)) {
                var ext = RegExp.$1;
                var mime
                = Components.classes ["@mozilla.org/mime;1"]
                .getService
                (Components.interfaces.nsIMIMEService);
                this.filetype
                = mime.getTypeFromExtension (ext);
            }
            else {
                this.filetype = "application/octet-stream";
            }
        }
        
        if (this.isCSS) {
            this.importFiles
            = UnMHTSave.checkCSS (this.originalContent,
                                  this.base || this.location,
                                  true,
                                  this.depth);
        }
        
        if (!this.notEncode
            && !this.isText) {
            if (this.encodedIndex > 0) {
                var data
                = this.originalContent.substr (this.encodedIndex);
                this.content += btoa (data);
            }
            else {
                this.content = btoa (this.originalContent);
            }
        }
        
        this.channel = null;
        this.httpChannel = null;
        this.status = UnMHTSave.STATUS_OK;
        
        if (this.isHTML) {
            if (UnMHTSave.enableInsertURL) {
                this.originalContent
                    = UnMHTSave.insertComment (this.originalContent,
                                               this.location);
            }
            if (UnMHTSave.enableInsertLink) {
                this.originalContent
                    = UnMHTSave.insertLink (this.originalContent,
                                            this.location);
            }
        }
        
        if (!this.notEncode
            && this.isText) {
            var bom = "";
            if (this.charset.match (/utf-8/i)) {
                if (!this.originalContent.match (/\xEF\xBB\xBF/)) {
                    bom = "=EF=BB=BF";
                }
            }
            this.content
            = bom + UnMHTConverter.encode_qp (this.originalContent);
        }
    },
    
    /**
     * データ到着のイベント
     *   nsIStreamListener.onDataAvailable
     *
     * @param  nsIRequest request
     *         対象のリクエスト
     * @param  nsISupports context
     *         ユーザ定義
     * @param  nsIInputStream inputStream
     *         データを取得するストリーム
     * @param  PRUint32 offset
     *         データの位置
     * @param  PRUint32 count 
     *         データの長さ
     */
    onDataAvailable : function (request, context, inputStream, offset, count) {
        var bstream
        = Components.classes ["@mozilla.org/binaryinputstream;1"]
        .createInstance (Components.interfaces.nsIBinaryInputStream);
        bstream.setInputStream (inputStream);
        this.originalContent += bstream.readBytes (count);
        this.recvSize += count;
        if (!this.notEncode
            && !this.isText) {
            while (this.originalContent.length
                   > this.encodedIndex + 12000) {
                var data
                    = this.originalContent.substr (this.encodedIndex, 12000);
                this.encodedIndex += 12000;
                this.content += btoa (data);
            }
        }
    },
    
    /**
     * キャッシュエントリが使用可能になったイベント
     *   nsICacheListener.onCacheEntryAvailable
     *
     * @param  nsICacheEntryDescriptor descriptor
     *         キャッシュの情報
     * @param  nsCacheAccessMode accessGranted
     *         アクセス権限
     * @param  nsresult status
     *         不明
     */
    onCacheEntryAvailable : function (descriptor, accessGranted, status) {
        if (this.status != UnMHTSave.STATUS_CACHE) {
            return;
        }
        if (accessGranted) {
            var head = "";
            try {
                head = descriptor.getMetaDataElement ("response-head");
            }
            catch (e) {
            }
            
            if (head.match (/^HTTP\/[0-9\.]+ ([0-9]+) /)) {
                var code = RegExp.$1;
                if (code >= 400) {
                    this.status = UnMHTSave.STATUS_CACHE_NG;
                    
                    return;
                }
            }
            if (head.match (/(^|[\r\n])location[ \r\n\t]*:[ \r\n\t]*([^\r\n]+)/i)) {
                var url = RegExp.$2;
                this.importFiles = new Array ();
                this.importFiles.push (url);
                this.status = UnMHTSave.STATUS_CACHE_NG;
                
                return;
            }
            else if (head.match (/content-encoding[ \r\n\t]*:[ \r\n\t]*(compress|deflate|gzip|x-compress|x-gzip)/i)) {
                var encoding = (RegExp.$1).toLowerCase ();
                    
                try {
                    var converter
                        = Components.classes
                        ["@mozilla.org/streamconv;1?from="
                         + encoding + "&to=uncompressed"]
                        .createInstance
                        (Components.interfaces.nsIStreamConverter);
                    converter.asyncConvertData (encoding, "uncompressed",
                                                this, null);
                    
                    var listener
                        = converter.QueryInterface
                        (Components.interfaces.nsIStreamListener);
                    listener.onStartRequest (null, null);
                    var istream = descriptor.openInputStream (0);
                    listener.onDataAvailable
                        (null, null,
                         istream, 0, descriptor.dataSize);
                    istream.close ();
                    listener.onStopRequest (null, null,
                                            Components.results.NS_OK);
                    
                    return;
                }
                catch (e) {
                    this.originalContent = "";
                    this.content = "";
                    
                    this.status = UnMHTSave.STATUS_CACHE_NG;
                }
            }
            else {
                var istream = descriptor.openInputStream (0);
                var bstream
                = Components.classes ["@mozilla.org/binaryinputstream;1"]
                .createInstance (Components.interfaces.nsIBinaryInputStream);
                bstream.setInputStream (istream);
                this.originalContent
                = bstream.readBytes (descriptor.dataSize);
                bstream.close ();
                istream.close ();
            
                this.onStopRequest (null, null,
                                    Components.results.NS_OK);
            }
        }
        else {
            this.status = UnMHTSave.STATUS_CACHE_NG;
        }
    }
};

/**
 * ファイルの管理をする
 */
var UnMHTFileManager = {
    /**
     * パスを展開、連結する
     *
     * @param  String base
     *         親ディレクトリのパス
     * @param  String sub
     *         相対パス、もしくは絶対パス
     * @return String
     *         連結したパス
     */
    jointPath : function (base, sub) {
        var i;
        var path;
        var sep;
        
        var protocol;
        var host;
        
        if (base == null) {
            return sub;
        }
        
        /* セパレータを検出 */
        if (base.indexOf ("\\") != -1) {
            /* Windows のパスの場合 */
            sep = "\\";
        }
        else if (base.indexOf ("/") != -1) {
            /* Mac OS X、Linux のパス、URI の場合 */
            sep = "/";
        }
        else if (base.indexOf (":") != -1) {
            /* Mac OS 9 以前のパスの場合 */
            sep = ":";
        }
        else {
            /* 区切りが見付からない場合連結できないので
             * サブのみ返す */
            return sub;
        }
        
        if (base.match (/^(.+:\/\/)([^\/]+)(.*)$/)) {
            /* URI の場合 */
            protocol = RegExp.$1;
            host = RegExp.$2;
            base = RegExp.$3;
        }
        else if (base.match (/^(.+:)(.*)$/)) {
            /* Windows のネイティブなパスの場合 */
            protocol = RegExp.$1;
            host = "";
            base = RegExp.$2;
        }
        else {
            /* その他のパスの場合 */
            protocol = "";
            host = "";
        }
        
        if (sub.match (/^(.+:\/\/)([^\/]+)(.*)$/) || 
            sub.match (/^([A-Za-z]+:)(.*)$/)) {
            /* sub が URI、もしくは Windows のネイティブな絶対パスの場合 */
            return sub;
        }
        
        if (sub.substring (0, 1) == sep) {
            /* sub が絶対パスの場合 */
            base = "";
        }
        
        path = base + sep + sub;
        var sepRegExp
        = new RegExp ("[^"
                      + sep.replace (/(\\\[\]\(\)\^\,\.\{\}\|\?\!\-\*\+\$)/,
                                     "\\$1")
                      + "]+",
                      "g");
        var newNames = [];
        /* パスを解析するだけなので代入はしない */
        path.replace
        (sepRegExp,
         function (matched) {
            if (matched == ".") {
            }
            else if (matched == "..") {
                newNames.pop ();
            }
            else if (matched == "") {
            }
            else {
                newNames.push (matched);
            }
            return "";
        });
        
        path = "";
        for (i = 0; i < newNames.length; i ++) {
            path += sep;
            path += newNames [i];
        }
        
        path = protocol + host + path;
        
        return path;
    }
};

/**
 * 文字コード変換器
 */
var UnMHTConverter = {
    converter : null, /* nsIScriptableUnicodeConverter  文字コード変換器 */
    
    /**
     * 初期化
     */
    init : function () {
        UnMHTConverter.converter
        = Components.classes ["@mozilla.org/intl/scriptableunicodeconverter"]
        .getService (Components.interfaces.nsIScriptableUnicodeConverter);
    },
    
    /**
     * UTF-16 から ISO-2022-JP に変換する
     *
     * @param  String text
     *         UTF-16 の文字列
     * @return String
     *         ISO-2002-JP に変換した文字列
     */
    convertToISO2022JP : function (text) {
        UnMHTConverter.converter.charset = "iso-2022-jp";
        return UnMHTConverter.converter.ConvertFromUnicode (text);
    },
    
    /**
     * HTML に使えない文字をエスケープする
     * 
     * @param  String text
     *         エスケープする文字列
     * @return String
     *         エスケープした文字列
     */
    escapeEntity : function (text) {
        return text
        .replace (/&/g, "&amp;")
        .replace (/\"/g, "&quot;")
        .replace (/\'/g, "&#x27;")
        .replace (/</g, "&lt;")
        .replace (/>/g, "&gt;")
        .replace (/\xa0/g, "&nbsp;");
    },
    
    /**
     * Quoted-Printable にエンコードする
     *
     * @param  String text
     *         エンコードする文字列
     * @return String
     *         エンコードした文字列
     */
    encode_qp : function (text) {
        var qp = new Array ();
        var line = "";
        
        text
        .replace (/\r\n|\r|\n/g, "\n")
        .replace
        (/([^\t\n\x20\x21-\x3c\x3e-\x7e])|(\n)|(.)/g,
         function (matched, unsafe, ret, safe) {
            var t = "";
            if (unsafe) {
                var hex = unsafe.charCodeAt (0).toString (16).toUpperCase ();
                if (hex.length == 1) {
                    hex = "0" + hex;
                }
            
                t = "=" + hex;
            }
            else if (ret) {
                if (line.substr (line.length - 1) == " ") {
                    qp.push (line.substr (0, line.length - 1) + "=");
                    line = "=20";
                }
                else if (line.substr (line.length - 1) == "\t") {
                    qp.push (line.substr (0, line.length - 1) + "=");
                    line = "=09";
                }
                
                qp.push (line);
                line = "";
                return;
            }
            else if (safe) {
                t = safe;
            }
            
            if (line.length + t.length > 75) {
                qp.push (line + "=");
                line = t;
            }
            else {
                line += t;
            }
        });
        
        if (line) {
            qp.push (line);
            line = "";
        }
        
        text = qp.join ("\r\n");
        
        return text;
    },
    
    /**
     * HTML に使えない文字のエスケープを解除する
     * 
     * @param  String text
     *         エスケープを解除する文字列
     * @return String
     *         エスケープを解除した文字列
     */
    unescapeEntity : function (text) {
        return text
        .replace (/&gt;/g, ">")
        .replace (/&lt;/g, "<")
        .replace (/&quot;/g, "\"")
        .replace (/&#x27;/g, "\'")
        .replace (/&nbsp;/g, " ")
        .replace (/&amp;/g, "&");
    },
    
    /**
     * Unicode から変換する
     *
     * @param  String text
     *         変換する文字列
     * @param  String charset
     *         変換先の文字コード
     * @return String
     *         変換した文字列
     */
    convertFromUnicode : function (text, charset) {
        var text2;
            
        UnMHTConverter.converter.charset = charset;
            
        try {
            text2 = UnMHTConverter.converter.ConvertFromUnicode (text);
        }
        catch (e) {
            text2 = text;
        }
            
        return text2;
    },
    
    /**
     * Unicode に変換する
     *
     * @param  String text
     *         変換する文字列
     * @param  String charset
     *         変換元の文字コード
     * @return String
     *         変換した文字列
     */
    convertToUnicode : function (text, charset) {
        var text2;
        UnMHTConverter.converter.charset = charset;
            
        try {
            text2 = UnMHTConverter.converter.ConvertToUnicode (text);
        }
        catch (e) {
            text2 = text;
        }
            
        return text2;
    },
    
    /**
     * encoded word をデコードする
     *
     * @param  String text
     *         デコードする文字列
     * @param  Boolean utf16
     *         text が UTF-16 であるかどうか
     * @return String
     *         デコードした文字列
     */
    decode_ew : function (text, utf16) {
        if (text.match (/^=\?([^\?]+)\?([^\?])\?([^\?]+)\?=$/)) {
            var encodedCharset = RegExp.$1;
            var encodedType = RegExp.$2;
            var encodedText = RegExp.$3;
            
            if (encodedType == "Q") {
                encodedText
                    = encodedText.replace
                    (/=([A-Fa-f0-9][A-Fa-f0-9])/g,
                     function (dummy, matched) {
                        return String.fromCharCode ("0x" + matched);
                    });
            }
            else if (encodedType == "B") {
                encodedText = atob (encodedText);
            }
            
            return UnMHTConverter.convertToUnicode (encodedText,
                                                    encodedCharset);
        }
        
        if (utf16) {
            return text;
        }
        
        return UnMHTConverter.convertToUnicode (text, "shift_jis");
    },
    
    /**
     * Base64 をデコードする
     *
     * @param  String text
     *         デコードする文字列
     * @return String
     *         デコードした文字列
     */
    atob : function (text) {
        var b
        = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        
        var result = "";
        var c, part1, part2, part3, part4;

        for (var i = 0; i < text.length; i += 4) {
            part1 = b.indexOf (text.substr (i, 1));
            part2 = b.indexOf (text.substr (i + 1, 1));
            part3 = b.indexOf (text.substr (i + 2, 1));
            part4 = b.indexOf (text.substr (i + 3, 1));
            
            if (part1 != -1 && part2 != -1) {
                result += String.fromCharCode ((part1 << 2)
                                               | (part2 >> 4));
            }
            if (part2 != -1 && part3 != -1 && part3 != 64) {
                result += String.fromCharCode (((part2 & 0x0f) << 4)
                                               | (part3 >> 2));
            }
            if (part3 != -1 && part4 != -1 && part4 != 64) {
                result += String.fromCharCode (((part3 & 0x03) << 6)
                                               | (part4));
            }
        }
        
        return result;
    },
    
    /**
     * Base64 にエンコードする
     *
     * @param  String text
     *         エンコードする文字列
     * @return String
     *         エンコードした文字列
     */
    btoa8 : function (text) {
        var b
        = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        
        var result = "";
        var c;
        var part1, part2, part3, part4;
        
        for (var i = 0; i < text.length; i += 3) {
            c = text.charCodeAt (i) & 0xff;
            part1 = c >> 2;
            part2 = (c & 0x3) << 4;
            if (i + 1 == text.length) {
                part3 = 64;
                part4 = 64;
            }
            else {
                c = text.charCodeAt (i + 1) & 0xff;
                part2 |= (c & 0xf0) >> 4;
                part3 = (c & 0x0f) << 2;
                if (i + 2 == text.length) {
                    part4 = 64;
                }
                else {
                    c = text.charCodeAt (i + 2) & 0xff;
                    part3 |= (c & 0xc0) >> 6;
                    part4 = c & 0x3f;
                }
            }
            result
                += b.substr (part1, 1)
                + b.substr (part2, 1)
                + b.substr (part3, 1)
                + b.substr (part4, 1);
        }
        return result;
    },
    
    /**
     * UTF-16 を Base64 にエンコードする
     *
     * @param  String text
     *         エンコードする文字列
     * @return String
     *         エンコードした文字列
     */
    btoa16 : function (text) {
        var b
        = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        
        var result = "";
        var c;
        var part1a, part2a, part3a, part4a;
        var part1b, part2b, part3b, part4b;
        
        for (var i = 0; i < text.length; i += 3) {
            c = (text.charCodeAt (i) >> 8) & 0xff;
            part1a = c >> 2;
            part2a = (c & 0x3) << 4;
            c = text.charCodeAt (i) & 0xff;
            part2a |= (c & 0xf0) >> 4;
            part3a = (c & 0x0f) << 2;
            if (i + 1 == text.length) {
                part4a = 64;
                partab = -1;
            }
            else {
                c = (text.charCodeAt (i + 1) >> 8) & 0xff;
                part3a |= (c & 0xc0) >> 6;
                part4a = c & 0x3f;
                c = text.charCodeAt (i + 1) & 0xff;
                part1b = c >> 2;
                part2b = (c & 0x3) << 4;
                if (i + 2 == text.length) {
                    part3b = 64;
                    part4b = 64;
                }
                else {
                    c = (text.charCodeAt (i + 2) >> 8) & 0xff;
                    part2b |= (c & 0xf0) >> 4;
                    part3b = (c & 0x0f) << 2;
                    c = text.charCodeAt (i + 2) & 0xff;
                    part3b |= (c & 0xc0) >> 6;
                    part4b = c & 0x3f;
                }
            }
            
            result
                += b.substr (part1a, 1)
                + b.substr (part2a, 1)
                + b.substr (part3a, 1)
                + b.substr (part4a, 1);
            
            if (part1b != -1) {
                result
                    += b.substr (part1b, 1)
                    + b.substr (part2b, 1)
                    + b.substr (part3b, 1)
                    + b.substr (part4b, 1);
            }
        }
        return result;
    }
};

/**
 * 保存の本体
 */
var UnMHTSave = {
    isThunderbird : false, /* Boolean  Thunderbird か */
    
    copyMode : false,    /* Boolean  元のファイルをコピーするモードか */
    headerList : null,   /* Array  ヘッダ
                          *   [[String フィールド名, String フィールド値], ...] */
    
    /* ファイルの取得状況 */
    STATUS_NA : 0,       /* Number  未取得 */
    STATUS_CACHE : 1,    /* Number  キャッシュ開始 */
    STATUS_CACHE_NG : 2, /* Number  キャッシュ失敗 */
    STATUS_NET : 3,      /* Number  ネットワーク開始 */
    STATUS_OK : 4,       /* Number  取得完了 */
    STATUS_NG : 5,       /* Number  取得失敗 */
    
    messages : null,     /* nsIStringBundle  メッセージ */
    isCanceled : false,  /* Boolean  キャンセルされたか */ 
    isStopped : false,   /* Boolean  エラー終了したか */
    
    extraHeaders : null, /* Array  拡張ヘッダ */
    
    mode : 1,            /* Number  動作モード
                          *   1： 単体
                          *   2:  SaveMulti からの呼び出し */
    
    targetFile : null,          /* nsIFile  対象のファイル */
    subject : "",               /* String  ファイルの subject */
    location : "",              /* String  ルートドキュメントの URI */ 
    
    urls : null,                /* Object  収集した URL */
    files : null,               /* Array  収集したファイル
                                 *   [UnMHTSaveFile, ...] */
    progressLabel : null,       /* XULElement  取得状況のラベル */
    progressCurrent : null,     /* XULElement  現在取得しているファイル */
    progressCurrentSize : null, /* XULElement  現在取得しているファイルのサイズ */
    progress : null,            /* XULElement  取得状況のバー */
    failedLabel : null,         /* XULElement  失敗したファイルのラベル */
    failedLabel2 : null,        /* XULElement  失敗したファイルのラベル2 */
    failedLabel3 : null,        /* XULElement  失敗したファイルのラベル3 */
    failedList : null,          /* XULElement  失敗したファイル */
    dialog : null,              /* XULElement  ダイアログ */
    acceptButton : null,        /* XULElement  OK ボタン */
    cancelButton : null,        /* XULElement  キャンセルボタン */
    
    enableInsertURL : false,    /* Boolean  URL をコメントとして挿入 */
    enableInsertLink : false,   /* Boolean  URL をリンクとして挿入 */
    enableShowErrors : false,   /* Boolean  保存されなかったファイルを表示する */
    enableSkipAlert : false,    /* Boolean  CSS の階層が深い場合に
                                 *   警告を表示しない */
    skipAlertDefault : "save",  /* String  警告を表示しない場合の動作 */
    skipAlertValue : "",        /* String  現在の値
                                 *   "": 未設定
                                 *   "save": 保存する
                                 *   "ignore": 無視する */
    
    /**
     * 初期化処理
     */
    init : function () {
        var arg = window.arguments [0];
        
        var targetFile = arg [0];
        var targetDocument = arg [1];
        var useCurrentState = arg [2];
        var isScrapBook = arg [3];
        var scrapBookData = arg [4];
        
        var subject;
        if (isScrapBook) {
            subject = scrapBookData ["title"];
        }
        else {
            subject = targetDocument.title;
        }
        
        UnMHTSave.initWith (targetFile,
                            useCurrentState,
                            targetDocument.location.href,
                            subject,
                            document.getElementById ("unmhtSaveDialog"),
                            1);
        
        setTimeout (function () {
                UnMHTSave.addDocument
                    (targetDocument, targetDocument.location.href,
                     useCurrentState, isScrapBook, scrapBookData);
                
                if (UnMHTSave.files.length > 0) {
                    setTimeout (UnMHTSave.checkFiles,
                                100);
                }
            }, 100);
    },
    
    /**
     * 初期化処理
     *
     * @param  nsIFile targetFile
     *         対象のファイル,
     * @param  Boolean useCurrentState
     *         現在の状態を使用するか
     * @param  String location
     *         メールを保存する際のルートドキュメントの URI
     * @param  String subject
     *         ファイルの subject
     * @param  Window dialog
     *         ダイアログ
     * @param  Number mode
     *         動作モード
     *           1: 単体のファイル
     *           2: 複数のファイル
     */
    initWith : function (targetFile, useCurrentState,
                         location, subject, dialog, mode) {
        UnMHTSave.copyMode = false;
        UnMHTSave.headerList = null;
        UnMHTSave.location = location;
        UnMHTSave.subject = subject;
        
        UnMHTSave.extraHeaders = new Array ();
        
        try {
            var prefBranch
                = Components.classes ["@mozilla.org/preferences-service;1"]
                .getService (Components.interfaces.nsIPrefBranch);
            if (prefBranch.prefHasUserValue
                ("unmht.save.inserturl")) {
                UnMHTSave.enableInsertURL
                    = prefBranch.getBoolPref
                    ("unmht.save.inserturl");
            }
            else {
                UnMHTSave.enableInsertURL = false;
            }
            if (prefBranch.prefHasUserValue
                ("unmht.save.insertlink")) {
                UnMHTSave.enableInsertLink
                    = prefBranch.getBoolPref
                    ("unmht.save.insertlink");
            }
            else {
                UnMHTSave.enableInsertLink = false;
            }
            if (prefBranch.prefHasUserValue
                ("unmht.save.show_errors")) {
                UnMHTSave.enableShowErrors
                    = prefBranch.getBoolPref
                    ("unmht.save.show_errors");
            }
            else {
                UnMHTSave.enableShowErrors = true;
            }
            if (prefBranch.prefHasUserValue
                ("unmht.save.show_alert_on_deep")) {
                UnMHTSave.enableSkipAlert
                    = prefBranch.getBoolPref
                    ("unmht.save.skip_alert_on_deep");
            }
            else {
                UnMHTSave.enableSkipAlert = false;
            }
            if (prefBranch.prefHasUserValue
                ("unmht.save.show_alert_on_deep.default")) {
                var value
                    = prefBranch.getCharPref
                    ("unmht.save.skip_alert_on_deep.default");
                UnMHTSave.skipAlertDefault = unescape (value);
            }
            else {
                UnMHTSave.skipAlertDefault = "save";
            }
            if (UnMHTSave.enableSkipAlert) {
                UnMHTSave.skipAlertValue = UnMHTSave.skipAlertDefault;
            }
        }
        catch (e) {
        }
        
        try {
            var id
            = Components.classes ["@mozilla.org/xre/app-info;1"]
            .getService (Components.interfaces.nsIXULAppInfo).ID;
            if (id == "{3550f703-e582-4d05-9a08-453d09bdfdc6}") {
                /* Thunderbird */
                UnMHTSave.isThunderbird = true;
            }
        } catch (e) {}        
        
        UnMHTSave.dialog = dialog;
        UnMHTSave.mode = mode;
        
        UnMHTSave.messages
        = Components.classes ["@mozilla.org/intl/stringbundle;1"]
        .getService (Components.interfaces.nsIStringBundleService)
        .createBundle ("chrome://unmht/locale/unmht.properties");
        
        UnMHTConverter.init ();
        
        UnMHTSave.targetFile = targetFile;
        
        UnMHTSave.urls = new Object ();
        UnMHTSave.files = new Array ();
        
        UnMHTSave.acceptButton = UnMHTSave.dialog.getButton ("accept");
        UnMHTSave.cancelButton = UnMHTSave.dialog.getButton ("cancel");
        
        UnMHTSave.acceptButton.disabled = true;
        
        if (UnMHTSave.mode == 1) {
            if (useCurrentState) {
                document.title
                    = document.getElementById ("current_label").value;
            }
            else  {
                document.title
                    = document.getElementById ("original_label").value;
            }
        }
        
        UnMHTSave.progressLabel
        = document.getElementById ("progress_label");
        UnMHTSave.progressCurrent
        = document.getElementById ("progress_current");
        UnMHTSave.progressCurrentSize
        = document.getElementById ("progress_current_size");
        UnMHTSave.progress
        = document.getElementById ("progress");
        UnMHTSave.failedLabel
        = document.getElementById ("failed_label");
        UnMHTSave.failedLabel2
        = document.getElementById ("failed_label2");
        UnMHTSave.failedLabel3
        = document.getElementById ("failed_label3");
        UnMHTSave.failedList
        = document.getElementById ("failed");
        
        UnMHTSave.progressLabel.value
        = UnMHTSave.messages.GetStringFromName ("save_begin");
        UnMHTSave.progress.setAttribute ("value", "0%");
    },
    
    /**
     * 拡張ヘッダを追加する
     *
     * @param  String name
     *         フィールド名
     * @param  String vlaue
     *         フィールド値
     */
    addExtraHeader : function (name, value) {
        UnMHTSave.extraHeaders.push (new Array (name, value));
    },
    
    /**
     * URL を収集する
     *
     * @param  HTMLDocument currentDocument
     *         対象のドキュメント
     * @param  Boolean useCurrentState
     *         現在の状態を使用するか
     * @param  HTMLElement node
     *         対象の要素
     * @param  String base
     *         ベースディレクトリ
     * @param  Number depth
     *         現在の深さ
     */
    getNormalDocumentFiles : function (currentDocument, location,
                                       useCurrentState,
                                       node, base, depth) {
        var url;
        var style;
        var file;
        var content;
        
        style = null;
        try {
            style = currentDocument.defaultView.getComputedStyle (node, "");
        }
        catch (e) {
        }
        if (style) {
            url = style.getPropertyCSSValue ("background-image");
            if (url && url.primitiveType == CSSPrimitiveValue.CSS_URI) {
                url = url.getStringValue ();
                url = UnMHTFileManager.jointPath (base, url);
                
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    file.status = UnMHTSave.STATUS_NA;
                    file.isCSS = false;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = "";
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
        }
        
        if (node instanceof Components.interfaces.nsIDOMHTMLStyleElement) {
            content = UnMHTSave.convertToPlainText (node);
            UnMHTSave.checkCSS
            (content, base, false, depth);
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLImageElement) {
            url = node.src;
            if (url) {
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    file.status = UnMHTSave.STATUS_NA;
                    file.isCSS = false;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = "";
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLLinkElement
                 && node.rel) {
            if (node.rel.match (/\bicon\b/i)) {
                url = node.href;
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = false;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
            }
            else if (node.rel.match (/\bstylesheet\b/i)) {
                url = node.href;
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = true;
                        file.isText = true;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "text/css";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLInputElement
                 && node.type.toLowerCase () == "image") {
            url = node.src;
            if (url) {
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    file.status = UnMHTSave.STATUS_NA;
                    file.isCSS = false;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = "";
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
        }
        /*
        else if (node instanceof Components.interfaces.nsIDOMHTMLAppletElement) {
            url = node.archive;
            if (node.hasAttribute ("codebase")
                && node.getAttribute ("codebase")) {
                url
                    = UnMHTFileManager.jointPath
                    (base, node.getAttribute ("codebase")) + "/";
            }
            url = UnMHTFileManager.jointPath (base, url);
            if (url) {
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    file.status = UnMHTSave.STATUS_NA;
                    file.isCSS = false;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = "";
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
            else {
                url = node.code;
                url = UnMHTFileManager.jointPath (base, url);
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = false;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
                url = node.object;
                url = UnMHTFileManager.jointPath (base, url);
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = false;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLObjectElement) {
            if (node.type
                && node.type.toLowerCase () == "text/html") {
                try {
                    UnMHTSave.getNormalDocument
                    (node.contentDocument,
                    node.contentDocument.location.href, useCurrentState, depth,
                    false);
                }
                catch (e) {
                }
            }
            else {
                url = node.data;
                url = UnMHTFileManager.jointPath (base, url);
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = false;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLParamElement) {
            if ("name" in node
                && node.name.toLowerCase () == "movie") {
                url = node.value;
                url = UnMHTFileManager.jointPath (base, url);
                if (url) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                    
                        file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = false;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
            }
        }
        */
        else if (node instanceof Components.interfaces.nsIDOMHTMLScriptElement) {
            url = node.src;
            if (url) {
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    file.status = UnMHTSave.STATUS_NA;
                    file.isCSS = false;
                    file.isText = true;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = "text/javascript";
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLEmbedElement) {
            url = node.src;
            url = UnMHTFileManager.jointPath (base, url);
            if (url) {
                if (!(url in UnMHTSave.urls)) {
                    UnMHTSave.urls [url] = 1;
                    
                    file = new UnMHTSaveFile (url);
                    file.depth = depth + 1;
                    
                    var ignoreFiles = [
                        "http://69.9.170.242/doubleagent.swf",
                        "http://admin.brightcove.com/viewer/",
                        "http://cdn.tvs.npowr.com/tvsplayer/Player.swf",
                        "http://l.yimg.com/cosmos.bcst.yahoo.com/",
                        "http://media1.break.com/static/app/v1/global/swf/ghost_player.swf",
                        "http://i.current.com/swf/barca/vm2_cdc.swf",
                        "http://lads.myspace.com/videos/vplayer.swf",
                        "http://raguun.jp/flv",
                        "http://s.mcstatic.com/f/fvp/",
                        "http://static.vsocial.com/flash/",
                        "http://video.google.com/googleplayer.swf",
                        "http://video.stage6.com/",
                        "http://videos.streetfire.net/flvplayer3.swf",
                        "http://vm2-2.vision.ameba.jp/mcs.swf",
                        "http://www.blastro.com/flashplayer/flvPlayer.swf",
                        "http://www.dailymotion.com/swf/",
                        "http://www.danerd.com/web_player.swf",
                        "http://www.guba.com/f/root.swf",
                        "http://www.lulu.tv/wp-content/flash_play/flvplayer-square-beta2.swf",
                        "http://www.musicvideocodes.info/asx.php",
                        "http://www.youtube.com/v/",
                        "/",
                        "."
                        ];
                    var ok = true;
                    
                    for (var k = 0; k < ignoreFiles.length; k ++) {
                        if (url.indexOf (ignoreFiles [k]) == 0) {
                            ok = false;
                            break;
                        }
                    }
                    
                    if (ok) {
                        file.status = UnMHTSave.STATUS_NA;
                    }
                    else {
                        file.status = UnMHTSave.STATUS_NG;
                    }
                    file.isCSS = false;
                    file.originalContent = "";
                    file.content = "";
                    file.filetype = node.getAttribute ("type");
                    file.charset = "";
                    UnMHTSave.files.push (file);
                }
            }
        }
        else if (node instanceof Components.interfaces.nsIDOMHTMLFrameElement
            || node instanceof Components.interfaces.nsIDOMHTMLIFrameElement) {
            try {
                UnMHTSave.getNormalDocument
                (node.contentDocument,
                 node.contentDocument.location.href, useCurrentState, depth,
                 false);
            }
            catch (e) {
            }
        }
    },
    
    /**
     * HTML ツリーを文字列に変換する
     *
     * @param  HTMLElement element
     *         対象の要素
     * @return String
     *         要素全体の HTML
     */
    convertToPlainText : function (element) {
        var text = "";
        var attributes = "";
        var name, value;
        var node = null;
        
        if (element
            && element.nodeName) {
            var nodeName = element.nodeName.toLowerCase ();
            
            if (nodeName == "#text") {
                text += UnMHTConverter.escapeEntity (element.nodeValue);
            }
            else if (nodeName == "#comment") {
                /* コメントをコメントとして展開する */
                text
                    += "<!--"
                    + element.nodeValue
                    + "-->";
            }
            else if (nodeName == "base"
                     && UnMHTSave.isThunderbird) {
            }
            else {
                if (element.attributes) {
                    for (var i = 0; i < element.attributes.length; i ++) {
                        name
                            = element.attributes.item (i).nodeName
                            .toLowerCase ();
                        value = element.attributes.item (i).nodeValue;
                        
                        if ((nodeName == "frame"
                             || nodeName == "iframe")
                            && name == "src") {
                            value = element.contentDocument.location.href;
                        }
                        
                        if ((name == "src" || name == "href")
                            && value.match
                            (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
                            value
                                = value.replace
                                (UnMHTSave.location,
                                 "file:///mail/attachment/");
                            if (value == "file:///mail/attachment/") {
                                value = "file:///mail/index.html";
                            }
                        }
                        
                        value = 
                            value
                            .replace (/&/g, "&amp;")
                            .replace (/\"/g, "&quot;")
                            .replace (/</g, "&lt;")
                            .replace (/>/g, "&gt;");
                        
                        attributes += " " + name +  "=" + "\"" + value + "\"";
                    }
                }
                
                text += "<" + nodeName + attributes + ">";
                
                if (nodeName == "noscript"
                    && element.firstChild
                    && element.firstChild.nodeName.toLowerCase () == "#text"
                    && element.firstChild.nextSibling == null) {
                    /* NOSCRIPT 要素が展開されていない場合 */
                    text += element.firstChild.nodeValue;
                }
                else if (nodeName == "noembed"
                    && element.firstChild
                    && element.firstChild.nodeName.toLowerCase () == "#text"
                    && element.firstChild.nextSibling == null) {
                    /* NOEMBED 要素が展開されていない場合 */
                    text += element.firstChild.nodeValue;
                }
                else if (nodeName == "style"
                         || nodeName == "script") {
                    node = element.firstChild;
                    while (node) {
                        text
                            += UnMHTConverter.unescapeEntity (node.nodeValue);
                        
                        node = node.nextSibling;
                    }
                }
                else {
                    node = element.firstChild;
                    while (node) {
                        text
                            += UnMHTSave.convertToPlainText (node);
                        
                        node = node.nextSibling;
                    }
                }
                
                if (nodeName != "br"
                    && nodeName != "img"
                    && nodeName != "imput"
                    && nodeName != "hr"
                    && nodeName != "meta") {
                    text += "</" + nodeName + ">";
                }
            }
        }
        
        return text;
    },
    
    /**
     * ドキュメントのデータを収集する
     *
     * @param  HTMLDocument currentDocument
     *         対象のドキュメント
     * @param  String location
     *         URL
     * @param  Boolean useCurrentState
     *         現在の状態を使用するか
     * @param  Number depth
     *         現在の深さ
     * @param  Boolean first
     *         最初のファイルか
     */
    getNormalDocument : function (currentDocument, location,
                                  useCurrentState, depth, first) {
        UnMHTSave.urls [location] = 1;
        
        var file = new UnMHTSaveFile (location);
        file.depth = depth + 1;
        var nodes;
        var i;
        
        if (UnMHTSave.mode == 1
            && first && file.extractor) {
            UnMHTSave.headerList = file.extractor.firstFile.headerList;
        }
        
        var base = location;
        var baseElement = null;
        if (base.match (/^(.+\/)([^\/]+)$/)) {
            base = RegExp.$1;
        }
        
        nodes = currentDocument.getElementsByTagName ("base");
        for (i = 0; i < nodes.length; i ++) {
            if (nodes [i].hasAttribute ("href")
                && nodes [i].getAttribute ("href")) {
                base
                    = UnMHTFileManager.jointPath
                    (base, nodes [i].getAttribute ("href")) + "/";
            }
        }
        
        file.base = base;
        
        if (useCurrentState) {
            file.originalContent = "";
            if (currentDocument.doctype) {
                file.originalContent
                    += "<!DOCTYPE "
                    + currentDocument.doctype.name
                    + " PUBLIC \""
                    + UnMHTConverter.escapeEntity
                    (currentDocument.doctype.publicId)
                    + "\" \""
                    + UnMHTConverter.escapeEntity
                    (currentDocument.doctype.systemId)
                    + "\">\n";
            }
            
            var head = null;
            var meta = null;
            var contentTypeExists = false;
            nodes = currentDocument.getElementsByTagName ("head");
            if (nodes.length > 0) {
                head = nodes [0];
            }
            nodes = head.getElementsByTagName ("meta");
            for (i = 0; i < nodes.length; i ++) {
                if (nodes [i].getAttribute ("http-equiv")
                    && nodes [i].getAttribute ("http-equiv")
                    .match (/Content-Type/i)
                    && nodes [i].getAttribute ("content")
                    && nodes [i].getAttribute ("content")
                    .match (/charset=/i)) {
                    contentTypeExists = true;
                    break;
                }
            }
            if (!contentTypeExists && currentDocument.inputEncoding) {
                meta = currentDocument.createElement ("meta");
                meta.setAttribute ("http-equiv", "Content-Type");
                meta.setAttribute ("content",
                                   "text/html; charset="
                                   + currentDocument.inputEncoding
                                   .toLowerCase ());
            }
            
            var tmpBaseElement = null;
            if (!baseElement) {
                tmpBaseElement = currentDocument.createElement ("base");
                tmpBaseElement.setAttribute ("href", base);
            }
            
            if (meta) {
                if (head.firstChild) {
                    head.insertBefore (meta, head.firstChild);
                }
                else {
                    head.appendChild (meta);
                }
            }
            if (tmpBaseElement) {
                head.appendChild (tmpBaseElement);
            }
            file.originalContent
                += UnMHTSave.convertToPlainText
                (currentDocument.documentElement);
            if (tmpBaseElement) {
                head.removeChild (tmpBaseElement);
            }
            if (meta) {
                head.removeChild (meta);
            }
            
            if (currentDocument.inputEncoding.match (/iso-8859/i)
                || currentDocument.inputEncoding.match (/us-ascii/i)
                || currentDocument.inputEncoding.match (/iso646-jp/i)
                || currentDocument.inputEncoding.match (/windows-1252/i)
                || currentDocument.inputEncoding.match (/macintosh/i)
                || currentDocument.inputEncoding.match (/koi8-r/i)
                || currentDocument.inputEncoding.match (/koi8-u/i)
                || currentDocument.inputEncoding.match (/ebcdic/i)) {
                file.originalContent
                    = file.originalContent.replace
                    (/[\u0100-\uffff]/g,
                     function (matched) {
                        return "&#" + matched.charCodeAt (0).toString () + ";";
                    });
            }
            
            if (UnMHTSave.enableInsertURL) {
                file.originalContent
                    = UnMHTSave.insertComment (file.originalContent,
                                               file.location);
            }
            if (UnMHTSave.enableInsertLink) {
                file.originalContent
                    = UnMHTSave.insertLink (file.originalContent,
                                            file.location);
            }
            
            file.status = UnMHTSave.STATUS_OK;
            file.isCSS = true;
            file.isHTML = true;
            file.isText = true;
            file.originalContent
                = UnMHTConverter.convertFromUnicode
                (file.originalContent,
                 currentDocument.inputEncoding);
            var bom = "";
            if (currentDocument.inputEncoding.match (/utf-8/i)) {
                bom = "=EF=BB=BF";
            }
            file.content
                = bom + UnMHTConverter.encode_qp (file.originalContent);
            file.filetype = "text/html";
            file.charset = currentDocument.inputEncoding;
        }
        else {
            file.status = UnMHTSave.STATUS_NA;
            file.isCSS = true;
            file.isHTML = true;
            file.isText = true;
            file.originalContent = "";
            file.content = "";
            file.filetype = "text/html";
            file.charset = currentDocument.inputEncoding;
        }
        UnMHTSave.files.push (file);
        
        var node;
        var firstNode;
        node = currentDocument.documentElement;
        firstNode = node;
        
        while (node) {
            UnMHTSave.getNormalDocumentFiles
                (currentDocument, location, useCurrentState, node, base, depth);
            
            if (node.firstChild) {
                node = node.firstChild;
            }
            else {
                while (node) {
                    if (node.nextSibling) {
                        node = node.nextSibling;
                        break;
                    }
                    node = node.parentNode;
                    if (node == firstNode) {
                        return;
                    }
                }
            }
        }
    },

    /**
     * URL をコメントとして挿入する
     *
     * @param  String content
     *         ファイルの内容
     * @param  String location
     *         URL
     * @return String
     *         ファイルの内容
     */
    insertComment : function (content, location) {
        var comment;
        comment = "<!-- Saved from " + location + " -->";
        if (content.match (/<html[^>]*>/i)) {
            content
                = content.replace
                (/<html[^>]*>/i,
                 function (matched) {
                    return matched + comment;
                });
        }
        else if (content.match (/<head[^>]*>/i)) {
            content
            = content.replace
            (/<head[^>]*>/i,
             function (matched) {
                return matched + comment;
            });
        }
        else if (content.match (/<body[^>]*>/i)) {
            content
            = content.replace
            (/<body[^>]*>/i,
             function (matched) {
                return matched + comment;
            });
        }
        else {
            content = comment + content;
        }
        
        return content;
    },

    /**
     * URL をリンクとして挿入する
     *
     * @param  String content
     *         ファイルの内容
     * @param  String location
     *         URL
     * @return String
     *         ファイルの内容
     */
    insertLink : function (content, location) {
        var link;
        link
        = "<div style=\"font-size: 0.8em; clear: both; margin-top: 2em; border: 1px solid black; background-color: white; color: black; padding: 0.5em; text-align: center;\">"
        + "Saved from <a class=\"unmht_link_to_original\" style=\"color: blue; background-color: inherit; text-decoration: underline;\" href=\"" + UnMHTConverter.escapeEntity (location) + "\">"
        + UnMHTConverter.escapeEntity (location)
        + "</a><br />"
        + (new Date ())
        + "<br />"
        + "Saved by UnMHT Ver." + UnMHTVersion
        + "</div>";
        if (content.match (/<\/body>/i)) {
            content
                = content.replace
                (/<\/body>/i,
                 function (matched) {
                    return link + matched;
                });
        }
        
        return content;
    },

    /**
     * CSS の import をチェックする
     *
     * @param  String content
     *         ファイルの内容
     * @param  String base
     *         CSS の記述されたファイル
     * @param  Boolean retUrls
     *         URL を返すか
     *           true:  URL を返す
     *           false:  URL を urls に挿入する
     * @param  Number depth
     *         現在の深さ
     * @return Array
     *         収集した URL
     */
    checkCSS : function (content, base, retUrls, depth) {
        var newUrls = new Array ();
        
        content.replace
        (/@import[ \t\r\n]+(url[ \t\r\n]*\([ \t\r\n]*)?[\"\']?([^\"\'\)]+)[\"\']?([ \t\r\n]*\))?/ig,
         function (matched, pre, url, post) {
            if (url) {
                if (base.match (/^(.+\/)([^\/]+)$/)) {
                    base = RegExp.$1;
                }
                url = UnMHTFileManager.jointPath (base, url);
                
                if (!retUrls) {
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                        
                        var file = new UnMHTSaveFile (url);
                        file.depth = depth + 1;
                        file.status = UnMHTSave.STATUS_NA;
                        file.isCSS = true;
                        file.isText = true;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "text/css";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
                else {
                    newUrls.push (url);
                }
            }
        });
        
        return newUrls;
    },
    
    /**
     * ドキュメントを追加する
     *
     * @param  HTMLDocument targetDocument
     *         対象のドキュメント
     * @param  String location
     *         URL
     * @param  Boolean useCurrentState
     *         現在の状態を使用するか
     * @param  Boolean isScrapBook
     *         ScrapBook か
     * @param  Object scrapBookData
     *         ScrapBook のデータ
     */
    addDocument : function (targetDocument,
                            location,
                            useCurrentState,
                            isScrapBook,
                            scrapBookData) {
        if (UnMHTSave.isCanceled) {
            UnMHTSave.stop ();
            return;
        }
        
        var istream;
        var content;
        
        try {
            var ios
                = Components.classes ["@mozilla.org/network/io-service;1"]
                .getService (Components.interfaces.nsIIOService);
            
            if (!isScrapBook
                && !useCurrentState
                && UnMHTSave.protocolHandler.isUnMHTURI (location)) {
                UnMHTSave.getUnMHTDocument (location);
            }
            else {
                if (isScrapBook) {
                    UnMHTSave.getScrapBook (scrapBookData);
                }
                else {
                    UnMHTSave.getNormalDocument
                        (targetDocument, location, useCurrentState, 0,
                         true);
                }
            }
        }
        catch (e) {
            UnMHTSave.progressLabel.value
            = UnMHTSave.messages.GetStringFromName ("error_save");
            UnMHTSave.progressCurrent.value
            = e;
            UnMHTSave.isStopped = true;
            if (UnMHTSave.mode == 2) {
                UnMHTSaveMulti.isStopped = true;
            }
        }
    },
    
    /**
     * データドキュメントを追加する
     *
     * @param  String content
     *         内容
     * @param  String location
     *         Content-Location の値
     * @param  String charset
     *         Content-Type のうち、文字コード
     * @param  String filetype
     *         Content-Type のうち、MIME タイプ
     */
    addDataDocument : function (content, location, charset, filetype) {
        var file = new UnMHTSaveFile (location);
        file.status = UnMHTSave.STATUS_OK;
        file.depth = 0;
        file.charset = charset;
                    
        file.filetype = filetype;
        if (file.filetype == "text/html") {
            file.isCSS = false;
            file.isHTML = true;
            file.isText = true;
        }
        else if (file.filetype == "text/css") {
            file.isCSS = false;
            file.isHTML = false;
            file.isText = true;
        }
        else if (file.filetype == "text/javascript") {
            file.isCSS = false;
            file.isHTML = false;
            file.isText = true;
        }
        else {
            file.isCSS = false;
            file.isHTML = false;
            file.isText = false;
        }
        
        file.originalContent = content;
        if (file.isText) {
            file.content
                = UnMHTConverter.encode_qp (file.originalContent);
        }
        else {
            file.content = btoa (file.originalContent);
        }
                    
        UnMHTSave.files.unshift (file);
    },
    
    /**
     * unmht:// なドキュメントのデータを収集する
     *
     * @param  HTMLDocument targetDocument
     *         対象のドキュメント
     */
    getUnMHTDocument : function (location) {
        var param = UnMHTSave.protocolHandler.getUnMHTURIParam (location);
        
        var entries
        = Components.classes ["@mozilla.org/appshell/window-mediator;1"]
        .getService (Components.interfaces.nsIWindowMediator)
        .getEnumerator ("navigator:browser");
        
        var original = param.original;
        var refName = param.refName;
        var baseURI = UnMHTSave.protocolHandler.enUnMHTURI (original);
        
        while (entries.hasMoreElements ()) {
            var w = entries.getNext ();
                    
            if (original in w.UnMHT.extractors) {
                var extractor = w.UnMHT.extractors [original];
                
                var srcFile = null;
                var rootFile;
                
                if (refName
                    && refName in extractor.refFiles) {
                    rootFile = extractor.refFiles [refName];
                }
                else {
                    rootFile = extractor.rootFile;
                }
                
                if (UnMHTSave.mode == 1
                    && rootFile == extractor.rootFile
                    && !UnMHTSave.enableInsertURL
                    && !UnMHTSave.enableInsertLink) {
                    // 元のファイルのコピー
                    UnMHTSave.copyMode = true;
                    
                    var file = new UnMHTSaveFile (original);
                    file.status = UnMHTSave.STATUS_NA;
                    file.depth = 0;
                    file.notEncode = true;
                    file.isCSS = false;
                    file.isHTML = false;
                    file.isText = false;
                    alert (original);
                    
                    UnMHTSave.files.push (file);
                }
                else {
                    if (UnMHTSave.mode == 1) {
                        UnMHTSave.headerList = extractor.firstFile.headerList;
                    }
                    
                    // 指定されたファイルを先頭にする
                    for (i = -1; i < extractor.files.length; i ++) {
                        var srcFile;
                        if (i == -1) {
                            srcFile = rootFile;
                        }
                        else {
                            srcFile = extractor.files [i];
                            if (srcFile == rootFile) {
                                continue;
                            }
                        }
                        if (srcFile == extractor.firstFile) {
                            UnMHTSave.headerList = null;
                        }
                        
                        var file = new UnMHTSaveFile (srcFile.location);
                        file.headerList = srcFile.headerList;
                        file.status = UnMHTSave.STATUS_OK;
                        file.depth = 0;
                        file.charset = srcFile.charset;
                    
                        file.filetype = srcFile.filetype;
                        if (file.filetype == "text/html") {
                            file.isCSS = false;
                            file.isHTML = true;
                            file.isText = true;
                        }
                        else if (file.filetype == "text/css") {
                            file.isCSS = false;
                            file.isHTML = false;
                            file.isText = true;
                        }
                        else if (file.filetype == "text/javascript") {
                            file.isCSS = false;
                            file.isHTML = false;
                            file.isText = true;
                        }
                        else {
                            file.isCSS = false;
                            file.isHTML = false;
                            file.isText = false;
                        }
                        var loc = baseURI;
                        if (srcFile != rootFile) {
                            loc += srcFile.refName;
                        }
                        
                        file.originalContent = srcFile.originalContent;
                        if (file.isHTML) {
                            if (UnMHTSave.enableInsertURL) {
                                file.originalContent
                                    = UnMHTSave.insertComment
                                    (file.originalContent,
                                     loc);
                            }
                            if (UnMHTSave.enableInsertLink) {
                                file.originalContent
                                    = UnMHTSave.insertLink
                                    (file.originalContent,
                                     loc);
                            }
                        }
                        if (file.isText) {
                            file.content
                                = UnMHTConverter.encode_qp
                                (file.originalContent);
                        }
                        else {
                            file.content = btoa (file.originalContent);
                        }
                    
                        UnMHTSave.files.push (file);
                    }
                }
                break;
            }
        }
    },
    
    /**
     * ScrapBook のデータを収集する
     *
     * @param  Object scrapBookData
     *         ScrapBook のデータ
     */
    getScrapBook : function (scrapBookData) {
        var dir = scrapBookData ["dir"].directoryEntries;
        while (dir.hasMoreElements ()) {
            var targetFile
                = dir.getNext ().QueryInterface (Components.interfaces.nsIFile);
            
            if (targetFile.leafName == "index.dat") {
                continue;
            }
            
            var isIndex = false;
            if (targetFile.leafName == "index.html") {
                isIndex = true;
            }
            
            var file = new UnMHTSaveFile (targetFile.leafName);
            file.depth = 0;
            file.status = UnMHTSave.STATUS_OK;
            file.isCSS = false;
            file.originalContent = "";
            file.filetype = "";
            file.charset = "";
            
            if (isIndex) {
                file.location = scrapBookData ["source"];
                file.filetype = "text/html";
                file.charset = scrapBookData ["charset"];
            }
            else {
                file.location
                    = UnMHTFileManager.jointPath
                    (scrapBookData ["source"],
                     file.location);
            }
            
            var fstream
            = Components
            .classes ["@mozilla.org/network/file-input-stream;1"]
            .createInstance (Components.interfaces.nsIFileInputStream);
            var bstream
            = Components.classes ["@mozilla.org/binaryinputstream;1"]
            .createInstance (Components.interfaces.nsIBinaryInputStream);
            fstream.init (targetFile, 0x01, 0444, 0);
            bstream.setInputStream (fstream);
            file.originalContent = bstream.readBytes (targetFile.fileSize);
            bstream.close ();
            fstream.close ();
            
            this.encodedIndex = 0;
            
            file.onStopRequest (null, null,
                                Components.results.NS_OK);
            
            if (UnMHTSave.enableInsertURL) {
                file.originalContent
                    = UnMHTSave.insertComment (file.originalContent,
                                               file.location);
            }
            if (UnMHTSave.enableInsertLink) {
                file.originalContent
                    = UnMHTSave.insertLink (file.originalContent,
                                            file.location);
            }
            
            if (isIndex) {
                UnMHTSave.files.unshift (file);
            }
            else {
                UnMHTSave.files.push (file);
            }
        }
    },
    
    /**
     * mht で保存 - ファイル収集のチェック
     */
    checkFiles : function () {
        if (UnMHTSave.isCanceled) {
            UnMHTSave.stop ();
            return;
        }
        
        var i;
        
        var ok, na, ng, start;
        ok = 0;
        na = 0;
        ng = 0;
        start = 0;
        
        var currentFile = "";
        var currentSize = "";
        
        var channel;
        var httpChannel;
        
        var now = (new Date ()).getTime ();
        
        var ios
        = Components.classes ["@mozilla.org/network/io-service;1"]
        .getService (Components.interfaces.nsIIOService);
        
        var cacheService
        = Components.classes ["@mozilla.org/network/cache-service;1"]
        .getService (Components.interfaces.nsICacheService);
        
        var httpCacheSession = null;
        for (var i = 0; i < UnMHTSave.files.length; i ++) {
            if (currentFile == ""
                && UnMHTSave.files [i].status == UnMHTSave.STATUS_NET
                && UnMHTSave.files [i].fileSize) {
                currentFile = UnMHTSave.files [i].location;
                currentSize
                    = UnMHTSave.files [i].recvSize + " bytes / "
                    + UnMHTSave.files [i].fileSize;
            }
            if (currentFile == ""
                && UnMHTSave.files [i].status != UnMHTSave.STATUS_OK
                && UnMHTSave.files [i].status != UnMHTSave.STATUS_NG) {
                currentFile = UnMHTSave.files [i].location;
                var size = UnMHTSave.files [i].fileSize;
                if (size == "") {
                    size = "unknown";
                }
                currentSize
                    = UnMHTSave.files [i].recvSize + " bytes / "
                    + size;
            }
            
            if ((UnMHTSave.files [i].status == UnMHTSave.STATUS_OK
                 || UnMHTSave.files [i].status == UnMHTSave.STATUS_NG)
                && UnMHTSave.files [i].importFiles
                && UnMHTSave.files [i].importFiles.length > 0) {
                var nextStatus = UnMHTSave.STATUS_NA;
                
                if (UnMHTSave.files [i].depth + 1 >= 10) {
                    if (UnMHTSave.skipAlertValue == "") {
                        UnMHTSave.getSkipAlertValue ();
                    }
                    
                    if (UnMHTSave.skipAlertValue == "ignore") {
                        nextStatus = UnMHTSave.STATUS_NG;
                    }
                }
                
                for (var j = 0;
                     j < UnMHTSave.files [i].importFiles.length; j ++) {
                    var url = UnMHTSave.files [i].importFiles [j];
                    if (!(url in UnMHTSave.urls)) {
                        UnMHTSave.urls [url] = 1;
                            
                        na ++;
                        var file = new UnMHTSaveFile (url);
                        file.depth = UnMHTSave.files [i].depth + 1;
                        file.status = nextStatus;
                        file.isText = true;
                        file.isCSS = true;
                        file.originalContent = "";
                        file.content = "";
                        file.filetype = "text/css";
                        file.charset = "";
                        UnMHTSave.files.push (file);
                    }
                }
                UnMHTSave.files [i].importFiles = null;
            }
            
            if (UnMHTSave.files [i].status == UnMHTSave.STATUS_NA) {
                na ++;
                UnMHTSave.files [i].startTime = (new Date ()).getTime ();
                UnMHTSave.files [i].status = UnMHTSave.STATUS_CACHE;
                try {
                    if (httpCacheSession == null) {
                        httpCacheSession
                            = cacheService.createSession
                            ("HTTP",
                             Components.interfaces.nsICache.STORE_ANYWHERE,
                             true);
                    }
                    
                    httpCacheSession
                        .asyncOpenCacheEntry
                        (UnMHTSave.files [i].location,
                         Components.interfaces.nsICache.ACCESS_READ,
                         UnMHTSave.files [i]);
                }
                catch (e) {
                    UnMHTSave.files [i].status = UnMHTSave.STATUS_CACHE_NG;
                }
                
            }
            else if (UnMHTSave.files [i].status == UnMHTSave.STATUS_CACHE) {
                na ++;
                if (now > UnMHTSave.files [i].startTime + 10 * 1000) {
                    UnMHTSave.files [i].status = UnMHTSave.STATUS_CACHE_NG;
                }
            }
            else if (UnMHTSave.files [i].status == UnMHTSave.STATUS_CACHE_NG) {
                UnMHTSave.files [i].startTime = (new Date ()).getTime ();
                UnMHTSave.files [i].status = UnMHTSave.STATUS_NET;
                na ++;
                start ++;
                try {
                    channel 
                        = ios.newChannel
                        (UnMHTSave.files [i].location, null, null);

                    UnMHTSave.files [i].channel = channel;
                    UnMHTSave.files [i].channel.asyncOpen
                        (UnMHTSave.files [i], null);
                }
                catch (e) {
                    UnMHTSave.files [i].status = UnMHTSave.STATUS_NG;
                }
                if (start > 32) {
                    break;
                }
            }
            else if (UnMHTSave.files [i].status == UnMHTSave.STATUS_NET) {
                na ++;
                if (now > UnMHTSave.files [i].startTime + 15 * 1000
                    && UnMHTSave.files [i].recvSize == 0) {
                    var NS_BINDING_ABORTED = 0x804b0002;
                    try {
                        UnMHTSave.files [i].channel.cancel (NS_BINDING_ABORTED);
                    }
                    catch (e) {
                    }
                    
                    UnMHTSave.files [i].status = UnMHTSave.STATUS_NG;
                }
            }
            else if (UnMHTSave.files [i].status == UnMHTSave.STATUS_OK) {
                ok ++;
            }
            else if (UnMHTSave.files [i].status == UnMHTSave.STATUS_NG) {
                ng ++;
            }
        }
        
        UnMHTSave.progressCurrent.value = currentFile;
        UnMHTSave.progressCurrentSize.value = currentSize;
        
        if (na > 0) {
            UnMHTSave.progressLabel.value
                = UnMHTSave.messages.GetStringFromName ("save_begin") + " "
                + (ok + ng) + "/" + (UnMHTSave.files.length);
            UnMHTSave.progress.setAttribute
                ("value",
                 parseInt (ok * 100 / UnMHTSave.files.length) + "%");
            
            setTimeout (UnMHTSave.checkFiles,
                        100);
        }
        else {
            UnMHTSave.progressLabel.value
            = UnMHTSave.messages.GetStringFromName ("save_begin") + " "
            + (ok + ng) + "/" + (UnMHTSave.files.length);
            UnMHTSave.progress.setAttribute ("value", "100%");
            
            setTimeout (UnMHTSave.storeToFile,
                        100);
        }
    },
    
    /**
     * CSS の階層が深い場合の警告を表示する
     */
    getSkipAlertValue : function () {
        var confirmURL = "chrome://unmht/content/confirm.xul";
        var features = "chrome,dialog,modal,centerscreen";
        /* arg
         *   0:result, 1:title, 2:message1, 3:message2,
         *   4:yes, 5:no, 6:abort, 7:rename, 8:applytoall,
         *   9:filename, 10:base, 11: existMessage1, 12: existMessage2,
         *   13:message3
         */
        var result = {
            result : 0,
            applyToAll: false,
            rename : false,
            filename : ""
        };
        
        openDialog (confirmURL, "", features,
                    [result,
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_title"),
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_message"),
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_message2"),
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_save"),
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_ignore"),
                     null,
                     null,
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_do_not_ask_again"),
                     null, null,
                     null, null,
                     UnMHTSave.messages.GetStringFromName
                     ("deep_css_message3")]);
        
        if (result.result == "yes") {
            UnMHTSave.skipAlertValue = "save";
        }
        else if (result.result == "no") {
            UnMHTSave.skipAlertValue = "ignore";
        }
        
        if (result.applyToAll) {
            try {
                var prefBranch
                = Components.classes ["@mozilla.org/preferences-service;1"]
                .getService (Components.interfaces.nsIPrefBranch);
                
                prefBranch.setBoolPref
                ("unmht.save.skip_alert_on_deep", true);
             
                UnMHTSave.skipAlertDefault = UnMHTSave.skipAlertValue;
                prefBranch.setCharPref
                ("unmht.save.skip_alert_on_deep.default",
                 escape (UnMHTSave.skipAlertDefault));
            }
            catch (e) {
            }
        }
    },
    
    /**
     * mht で保存 - ファイルの保存
     */
    storeToFile : function () {
        if (UnMHTSave.isCanceled) {
            UnMHTSave.stop ();
            return;
        }
        
        var i;
        
        var fstream;
        
        var fstream
        = Components
        .classes ["@mozilla.org/network/file-output-stream;1"]
        .createInstance (Components.interfaces.nsIFileOutputStream);
        try {
            fstream.init (UnMHTSave.targetFile, 0x02 | 0x08 | 0x20, 0664, 0);
        }
        catch (e) {
            UnMHTSave.progressLabel.value
            = UnMHTSave.messages.GetStringFromName ("error_open");
            UnMHTSave.isStopped = true;
            if (UnMHTSave.mode == 2) {
                UnMHTSaveMulti.isStopped = true;
            }
            return;
        }
        
        if (UnMHTSave.copyMode) {
            var file = UnMHTSave.files [0];
            fstream.write (file.originalContent, file.originalContent.length);
        }
        else {
            /* ヘッダに必要な情報を設定する */
            var subject
            = "=?iso-2022-jp?B?"
            + btoa (UnMHTConverter.convertToISO2022JP (UnMHTSave.subject))
            + "?=";
            var boundary = "----=_NextPart_000_0000_";
            var hex = [
                "0", "1", "2", "3", "4", "5", "6", "7",
                "8", "9", "A", "B", "C", "D", "E", "F"
                ];
            for (i = 0; i < 8; i ++) {
                boundary += hex [parseInt (Math.random () * 15)];
            }
            boundary += ".";
            for (i = 0; i < 8; i ++) {
                boundary += hex [parseInt (Math.random () * 15)];
            }
            var date
            = (new Date ()).toString ()
            .replace (/^([A-Za-z]{3}) /, "$1, ")
            .replace (/ \(.*\)$/, "");
        
            var data = "";
        
            /* 全体のヘッダ */
            data
            = "From: <Saved by UnMHT>\r\n"
            + "Subject: " + subject + "\r\n"
            + "Date: " + date + "\r\n"
            + "MIME-Version: 1.0\r\n"
            + "Content-Type: multipart/related;\r\n"
            + "\tboundary=\"" + boundary + "\";\r\n"
            + "\ttype=\"text/html\"\r\n";
            for (var i = 0; i < UnMHTSave.extraHeaders.length; i ++) {
                data += UnMHTSave.extraHeaders [i][0]
                    + ": " + UnMHTSave.extraHeaders [i][1] + "\r\n";
            }
            if (UnMHTSave.headerList) {
                for (var j = 0; j < UnMHTSave.headerList.length; j ++) {
                    var name = UnMHTSave.headerList [j][0];
                    var value = UnMHTSave.headerList [j][1];
                    if (name == "From"
                        || name == "Subject"
                        || name == "Date"
                        || name == "MIME-Version"
                        || name == "Content-Type") {
                        continue;
                    }
                    data += name + ": " + value + "\r\n";
                }
            }
            data
            += "\r\n";

            fstream.write (data, data.length);
        
            for (i = 0; i < UnMHTSave.files.length; i ++) {
                var file = UnMHTSave.files [i];
            
                if (file.status == UnMHTSave.STATUS_NG) {
                    continue;
                }
            
                data = "--" + boundary + "\r\n";
                fstream.write (data, data.length);
            
                data = "";
                if (file.charset) {
                    data += "Content-Type: " + file.filetype
                        + "; charset=\"" + file.charset + "\"\r\n";
                }
                else {
                    data += "Content-Type: " + file.filetype + "\r\n";
                }
                if (file.isText) {
                    data += "Content-Transfer-Encoding: quoted-printable\r\n";
                }
                else {
                    data += "Content-Transfer-Encoding: base64\r\n";
                }
            
                var location = file.location;
                if (location.search
                    (new RegExp ("[\u0100-\uffff]")) != -1) {
                    location
                        = "=?iso-2022-jp?B?"
                        + btoa (UnMHTConverter.convertToISO2022JP
                                (file.location)) + "?=";
                }
                if (location.match
                    (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
                    location
                        = location.replace (UnMHTSave.location,
                                            "file:///mail/attachment/");
                    if (location == "file:///mail/attachment/") {
                        location = "file:///mail/index.html";
                    }
                }
                data += "Content-Location: " + location + "\r\n";
                if (file.headerList) {
                    for (var j = 0; j < file.headerList.length; j ++) {
                        var name = file.headerList [j][0];
                        var value = file.headerList [j][1];
                        if (name == "Content-Location"
                            || name == "Content-Transfer-Encoding"
                            || name == "Content-Type") {
                            continue;
                        }
                        data += name + ": " + value + "\r\n";
                    }
                }
            
                data += "\r\n";
                fstream.write (data, data.length);
            
                if (file.isText) {
                    fstream.write (file.content,
                                   file.content.length);
                }
                else {
                    var width = 76;
                    data = "";
                    for (var j = 0; j < file.content.length; j += width) {
                        data
                            += file.content.substr (j, width) + "\r\n";
                    }
                    fstream.write (data,
                                   data.length);
                }
            
                data = "\r\n";
                fstream.write (data, data.length);
            }
        
            data = "--" + boundary + "--\r\n";
            fstream.write (data, data.length);
        }
        
        fstream.close ();
        
        UnMHTSave.progressLabel.value
        = UnMHTSave.messages.GetStringFromName ("save_done");
        
        var ngList = "";
        for (var i = 0; i < UnMHTSave.files.length; i ++) {
            if (UnMHTSave.files [i].status == UnMHTSave.STATUS_NG) {
                ngList += UnMHTSave.files [i].location + "\n";
            }
        }
        
        try {
            var dlManager
            = Components.classes ["@mozilla.org/download-manager;1"]
            .getService (Components.interfaces.nsIDownloadManager);
            var srcURI
            = Components.classes ["@mozilla.org/network/standard-url;1"]
            .createInstance (Components.interfaces.nsIURI);
            srcURI.spec = UnMHTSave.location;
            var ios = Components.classes ["@mozilla.org/network/io-service;1"].
            getService (Components.interfaces.nsIIOService);
            var targetURI
            = ios.newFileURI (UnMHTSave.targetFile);
            
            var mime = Components.classes ["@mozilla.org/mime;1"]
            .getService (Components.interfaces.nsIMIMEService);
            var mimeInfo
            = mime.getFromTypeAndExtension ("message/rfc822", "mht");
            
            var cancelable = {
                QueryInterface : function (iid) {
                    if (iid.equals (Components.interfaces.nsISupports)
                        || iid.equals (Components.interfaces
                                       .nsISupportsWeakReference)
                        || iid.equals (Components.interfaces.nsIStreamListener)
                        || iid.equals (Components.interfaces
                                       .nsIWebBrowserPersist)
                        || iid.equals (Components.interfaces.nsICancelable)) {
                        return this;
                    }
                    
                    throw Components.results.NS_NOINTERFACE;
                },
                
                cancel : function (reason) {
                }
            };
            
            var dl;
            var listener = new UnMHTSaveDownloadListener ();
            listener.fileSize = UnMHTSave.targetFile.fileSize;
            if (dlManager.addDownload.length == 7) {
                dl
                = dlManager.addDownload
                    (srcURI, targetURI,
                     UnMHTSave.targetFile.leafName,
                     null, null, null,
                     listener);
            }
            else if (dlManager.addDownload.length == 8) {
                dl
                = dlManager.addDownload
                    (0, srcURI, targetURI,
                     UnMHTSave.targetFile.leafName,
                     null, null, null,
                     listener);
            }
            else if (dlManager.addDownload.length == 9) {
                dl
                = dlManager.addDownload
                (0, srcURI, targetURI,
                 UnMHTSave.targetFile.leafName,
                 null, null, null, null,
                 listener);
            }
            else {
                dl
                = dlManager.addDownload
                (srcURI, targetURI,
                 UnMHTSave.subject,
                 null, null,
                 listener);
            }
            listener.dl = dl;
            
            var dummyURI
            = Components.classes ["@mozilla.org/network/standard-url;1"]
            .createInstance (Components.interfaces.nsIURI);
            dummyURI.spec = "unmht://dummy/hoge/fuga/";
            
            var channel = ios.newChannelFromURI (dummyURI);
            channel.asyncOpen (listener, null);
        }
        catch (e) {
        }
        
        if (UnMHTSave.enableShowErrors && ngList) {
            UnMHTSave.acceptButton.disabled = false;
            UnMHTSave.cancelButton.disabled = true;
            UnMHTSave.failedList.value += ngList;
            UnMHTSave.failedList.hidden = false;
            UnMHTSave.failedLabel.hidden = false;
            UnMHTSave.failedLabel2.hidden = false;
            UnMHTSave.failedLabel3.hidden = false;
            window.sizeToContent ();
            
            if (UnMHTSave.mode == 2) {
                UnMHTSaveMulti.done ();
            }
        }
        else {
            if (UnMHTSave.mode == 1) {
                setTimeout (function (window) {
                        window.close ();
                    }, 1000, window);
            }
            else {
                UnMHTSaveMulti.done ();
            }
        }
    },
    
    /**
     * キャンセルボタンのイベント
     */
    cancel : function () {
        if (UnMHTSave.isStopped) {
            return true;
        }
        
        UnMHTSave.isCanceled = true;
        
        try {
            for (var i = 0; i < UnMHTSave.files.length; i ++) {
                var NS_BINDING_ABORTED = 0x804b0002;
                try {
                    UnMHTSave.files [i].channel.cancel (NS_BINDING_ABORTED);
                }
                catch (e) {
                }
            }
        }
        catch (e) {
        }
        
        return false;
    },
    
    /**
     * 保存の停止
     */
    stop : function () {
        if (UnMHTSave.mode == 1) {
            window.close ();
        }
        else {
            UnMHTSaveMulti.done ();
        }
    }
};

var UnMHT_isThunderbird = false;
try {
    if (Components.classes ["@mozilla.org/xre/app-info;1"]
        .getService (Components.interfaces.nsIXULAppInfo).ID
        == "{3550f703-e582-4d05-9a08-453d09bdfdc6}") {
        UnMHT_isThunderbird = true;
    }
} catch (e) {}

if (UnMHT_isThunderbird) {
    /* Thunderbird */
    UnMHTSave.protocolHandler = {
        enUnMHTURI : function (uri) {
            return uri;
        },
        isUnMHTURI : function (uri) {
            return false;
        },
        deUnMHTURI : function (uri) {
            return uri;
        },
        getUnMHTURIParam : function (uri) {
            return null;
        }
    };
}
else {
    UnMHTSave.protocolHandler
        = Components.classes ["@mozilla.org/network/protocol;1?name=unmht"]
        .getService (Components.interfaces.arIUnMHTProtocolHandler);
}
